# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_array_almost_equal
from bluepysnap import Circuit
from utils import TEST_DATA_DIR

import connectome_manipulator.model_building.conn_props as test_module


def test_extract():
    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    src_mtypes = np.unique(nodes[0].get(src_ids, properties="mtype"))
    tgt_mtypes = np.unique(nodes[1].get(tgt_ids, properties="mtype"))
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)
    tab_mtypes_src = (
        nodes[0].get(edges_table["@source_node"].to_numpy(), properties="mtype").to_numpy()
    )
    tab_mtypes_tgt = (
        nodes[1].get(edges_table["@target_node"].to_numpy(), properties="mtype").to_numpy()
    )

    # Extract selected synapse/connection properties per pathway (= reference data)
    sel_prop = "conductance"
    nsyn_data = {
        k: np.full((len(src_mtypes), len(tgt_mtypes)), np.nan)
        for k in ["mean", "std", "min", "max"]
    }
    prop_data = {
        k: np.full((len(src_mtypes), len(tgt_mtypes), 1), np.nan)
        for k in ["mean", "std", "shared_within", "min", "max"]
    }
    for i1, mt1 in enumerate(src_mtypes):
        for i2, mt2 in enumerate(tgt_mtypes):
            tab_sel = edges_table.loc[np.logical_and(tab_mtypes_src == mt1, tab_mtypes_tgt == mt2)]
            if tab_sel.size == 0:
                continue
            _, syn_conn_idx, nsyn_conn = np.unique(
                tab_sel[["@source_node", "@target_node"]],
                axis=0,
                return_inverse=True,
                return_counts=True,
            )
            nsyn_data["mean"][i1, i2] = np.mean(nsyn_conn)
            nsyn_data["std"][i1, i2] = np.std(nsyn_conn)
            nsyn_data["min"][i1, i2] = np.min(nsyn_conn)
            nsyn_data["max"][i1, i2] = np.max(nsyn_conn)

            # Check if property value shared within connection
            if len(nsyn_conn) > 1 and len(np.unique(tab_sel[sel_prop])) == 1:
                # In case of a constant distribution, assume no sharing
                is_shared = False
            else:
                is_shared = True
                for cidx in range(len(nsyn_conn)):
                    if len(np.unique(tab_sel.loc[syn_conn_idx == cidx, sel_prop])) > 1:
                        # Found different property values within same connection
                        is_shared = False
                        break

            # Get property values
            prop_values = []
            for cidx in range(len(nsyn_conn)):
                if is_shared:
                    # Shared within connection, so take only first value
                    prop_values.append(tab_sel.loc[syn_conn_idx == cidx, sel_prop].iloc[0])
                else:
                    # Different values within connection, so take all values
                    prop_values.append(tab_sel.loc[syn_conn_idx == cidx, sel_prop].to_numpy())
            prop_values = np.hstack(prop_values)

            # Compute statistics
            prop_data["mean"][i1, i2] = np.mean(prop_values)
            prop_data["std"][i1, i2] = np.std(prop_values)
            prop_data["shared_within"][i1, i2] = is_shared
            prop_data["min"][i1, i2] = np.min(prop_values)
            prop_data["max"][i1, i2] = np.max(prop_values)

    # Check extraction (w/o histograms)
    res = test_module.extract(
        c,
        min_sample_size_per_group=None,
        max_sample_size_per_group=None,
        hist_bins=51,
        sel_props=[sel_prop],
        sel_src=None,
        sel_dest=None,
    )
    assert_array_equal(res["m_types"][0], src_mtypes)
    assert_array_equal(res["m_types"][1], tgt_mtypes)
    assert_array_equal(res["syn_props"], [sel_prop])
    for k in nsyn_data.keys():
        assert_array_equal(res["syns_per_conn_data"][k], nsyn_data[k])
    for k in prop_data.keys():
        assert_array_almost_equal(res["conn_prop_data"][k], prop_data[k])

    # Check extraction with Inf min sample size => Everything should be NaN
    res = test_module.extract(
        c,
        min_sample_size_per_group=np.inf,
        max_sample_size_per_group=None,
        hist_bins=50,
        sel_props=[sel_prop],
        sel_src=None,
        sel_dest=None,
    )
    assert_array_equal(res["m_types"][0], src_mtypes)
    assert_array_equal(res["m_types"][1], tgt_mtypes)
    assert_array_equal(res["syn_props"], [sel_prop])
    for k in nsyn_data.keys():
        assert_array_equal(res["syns_per_conn_data"][k], np.full_like(nsyn_data[k], np.nan))
    for k in prop_data.keys():
        assert_array_equal(res["conn_prop_data"][k], np.full_like(prop_data[k], np.nan))


def test_build():
    m_types = [["L4_MC", "L4_PC", "L5_PC"]] * 2
    m_type_class = [["INH", "EXC", "EXC"]] * 2
    m_type_layer = [[4, 4, 5]] * 2

    props = ["conductance", "u_syn"]
    src_mtypes = m_types[0]
    tgt_mtypes = m_types[1]
    np.random.seed(0)
    nsyn_data = {
        k: np.random.randint(low=1, high=10, size=(len(src_mtypes), len(tgt_mtypes)))
        for k in ["mean", "std", "min", "max", "norm_loc", "norm_scale"]
    }
    prop_data = {
        k: np.random.rand(len(src_mtypes), len(tgt_mtypes), len(props))
        for k in ["mean", "std", "shared_within", "min", "max", "norm_loc", "norm_scale"]
    }
    prop_data["shared_within"] = np.round(prop_data["shared_within"])

    # Check model building (default settings)
    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        props,
        distr_types={},
        data_types={},
        data_bounds={},
    )
    for sidx, s_mt in enumerate(src_mtypes):
        for tidx, t_mt in enumerate(tgt_mtypes):
            for pidx, pp in enumerate(props):
                model_distr = res.get_distr_props(prop_name=pp, src_type=s_mt, tgt_type=t_mt)
                assert model_distr["type"] == "normal"
                assert model_distr["mean"] == prop_data["mean"][sidx, tidx, pidx]
                assert model_distr["std"] == prop_data["std"][sidx, tidx, pidx]
                assert model_distr["shared_within"] == prop_data["shared_within"][sidx, tidx, pidx]
            model_distr = res.get_distr_props(
                prop_name="n_syn_per_conn", src_type=s_mt, tgt_type=t_mt
            )
            assert model_distr["mean"] == nsyn_data["mean"][sidx, tidx]
            assert model_distr["std"] == nsyn_data["std"][sidx, tidx]

    with pytest.raises(KeyError):
        res.get_distr_props(prop_name="WRONG_NAME", src_type=s_mt, tgt_type=t_mt)
    with pytest.raises(KeyError):
        res.get_distr_props(prop_name="n_syn_per_conn", src_type="WRONG_TYPE", tgt_type=t_mt)
    with pytest.raises(KeyError):
        res.get_distr_props(prop_name="n_syn_per_conn", src_type=s_mt, tgt_type="WRONG_TYPE")

    # Check distribution types (truncnorm)
    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        props,
        distr_types={pp: "truncnorm" for pp in props},
        data_types={},
        data_bounds={},
    )
    for sidx, s_mt in enumerate(src_mtypes):
        for tidx, t_mt in enumerate(tgt_mtypes):
            for pidx, pp in enumerate(props):
                model_distr = res.get_distr_props(prop_name=pp, src_type=s_mt, tgt_type=t_mt)
                assert model_distr["type"] == "truncnorm"
                assert model_distr["norm_loc"] == prop_data["norm_loc"][sidx, tidx, pidx]
                assert model_distr["norm_scale"] == prop_data["norm_scale"][sidx, tidx, pidx]
                assert model_distr["min"] == prop_data["min"][sidx, tidx, pidx]
                assert model_distr["max"] == prop_data["max"][sidx, tidx, pidx]
                assert model_distr["shared_within"] == prop_data["shared_within"][sidx, tidx, pidx]

    with pytest.raises(AssertionError, match='Distribution type "WRONG_TYPE" not supported!'):
        test_module.build(
            nsyn_data,
            prop_data,
            m_types,
            m_type_class,
            m_type_layer,
            props,
            distr_types={pp: "WRONG_TYPE" for pp in props},
            data_types={},
            data_bounds={},
        )
    with pytest.raises(AssertionError, match="ERROR: Not all required attribute"):
        test_module.build(
            nsyn_data,
            prop_data,
            m_types,
            m_type_class,
            m_type_layer,
            props,
            distr_types={pp: "discrete" for pp in props},
            data_types={},
            data_bounds={},
        )

    # Check data types
    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        props,
        distr_types={},
        data_types={pp: "int" for pp in props},
        data_bounds={},
    )
    for sidx, s_mt in enumerate(src_mtypes):
        for tidx, t_mt in enumerate(tgt_mtypes):
            for pidx, pp in enumerate(props):
                model_distr = res.get_distr_props(prop_name=pp, src_type=s_mt, tgt_type=t_mt)
                assert model_distr["dtype"] == "int"
                assert isinstance(res.draw(prop_name=pp, src_type=s_mt, tgt_type=t_mt)[0], np.int64)

    # Check data bounds
    test_bound = 10.0
    with pytest.raises(AssertionError, match="Data bounds error!"):
        test_module.build(
            nsyn_data,
            prop_data,
            m_types,
            m_type_class,
            m_type_layer,
            props,
            distr_types={},
            data_types={},
            data_bounds={pp: [2 * test_bound, test_bound] for pp in props},
        )

    res_l = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        props,
        distr_types={},
        data_types={},
        data_bounds={pp: [test_bound, np.inf] for pp in props},
    )
    res_u = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        props,
        distr_types={},
        data_types={},
        data_bounds={pp: [-np.inf, -test_bound] for pp in props},
    )
    for sidx, s_mt in enumerate(src_mtypes):
        for tidx, t_mt in enumerate(tgt_mtypes):
            for pidx, pp in enumerate(props):
                model_distr_l = res_l.get_distr_props(prop_name=pp, src_type=s_mt, tgt_type=t_mt)
                model_distr_u = res_u.get_distr_props(prop_name=pp, src_type=s_mt, tgt_type=t_mt)
                assert model_distr_l["lower_bound"] == test_bound
                assert model_distr_u["upper_bound"] == -test_bound
                assert res_l.draw(prop_name=pp, src_type=s_mt, tgt_type=t_mt)[0] == test_bound
                assert res_u.draw(prop_name=pp, src_type=s_mt, tgt_type=t_mt)[0] == -test_bound

    # Check synapse generation (w/o randomization)
    n_syn = 5.0
    p_val = 2.0
    prop_data = {
        k: np.full((len(src_mtypes), len(tgt_mtypes), len(props)), p_val)
        for k in ["mean", "min", "max"]
    }
    prop_data.update({k: np.zeros((len(src_mtypes), len(tgt_mtypes), len(props))) for k in ["std"]})
    prop_data.update(
        {k: np.ones((len(src_mtypes), len(tgt_mtypes), len(props))) for k in ["shared_within"]}
    )
    nsyn_data = {
        k: np.full((len(src_mtypes), len(tgt_mtypes)), n_syn) for k in ["mean", "min", "max"]
    }
    nsyn_data.update({"std": np.zeros((len(src_mtypes), len(tgt_mtypes)))})
    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        props,
        distr_types={},
        data_types={},
        data_bounds={},
    )
    for sidx, s_mt in enumerate(src_mtypes):
        for tidx, t_mt in enumerate(tgt_mtypes):
            syn = res.apply(src_type=s_mt, tgt_type=t_mt)
            assert syn.shape[0] == n_syn
            assert syn.shape[1] == len(props)
            assert np.all(syn == p_val)

    # Check interpolation
    m_types = [["L4_PC:A", "L4_PC:B", "L5_PC", "L5_MC"]] * 2
    m_type_class = [["EXC", "EXC", "EXC", "INH"]] * 2
    m_type_layer = [[4, 4, 5, 5]] * 2

    prop = "conductance"
    src_mtypes = m_types[0]
    tgt_mtypes = m_types[1]
    np.random.seed(0)
    nsyn_data = {
        k: np.random.randint(low=1, high=10, size=(len(src_mtypes), len(tgt_mtypes))).astype(float)
        for k in ["mean", "std", "min", "max", "norm_loc", "norm_scale"]
    }
    prop_data = {
        k: np.random.rand(len(src_mtypes), len(tgt_mtypes), 1)
        for k in ["mean", "std", "shared_within", "min", "max", "norm_loc", "norm_scale"]
    }
    prop_data["shared_within"] = np.round(prop_data["shared_within"])

    nsyn_data.update(
        {
            k: np.full((len(src_mtypes), len(tgt_mtypes)), np.nan, dtype=object)
            for k in ["val", "cnt", "p"]
        }
    )
    prop_data.update(
        {
            k: np.full((len(src_mtypes), len(tgt_mtypes), 1), np.nan, dtype=object)
            for k in ["val", "cnt", "p"]
        }
    )
    for sidx, s_mt in enumerate(src_mtypes):
        for tidx, t_mt in enumerate(tgt_mtypes):
            fake_data = np.random.randint(20, size=10)
            val, cnt = np.unique(fake_data, return_counts=True)
            nsyn_data["val"][sidx, tidx] = val
            nsyn_data["cnt"][sidx, tidx] = cnt
            nsyn_data["p"][sidx, tidx] = cnt / np.sum(cnt)

            fake_data = np.random.randint(20, size=10)
            val, cnt = np.unique(fake_data, return_counts=True)
            prop_data["val"][sidx, tidx, 0] = val
            prop_data["cnt"][sidx, tidx, 0] = cnt
            prop_data["p"][sidx, tidx, 0] = cnt / np.sum(cnt)

    test_src = "L4_PC:A"
    test_tgt = "L4_PC:A"
    test_src_idx = np.where(np.array(src_mtypes) == test_src)[0][0]
    test_tgt_idx = np.where(np.array(tgt_mtypes) == test_tgt)[0][0]

    def comb_discrete(vals, cnts):
        """Combines two discrete distributions. [Assuming no NaNs in vals/cnts!]"""
        comb = np.hstack(
            [np.hstack([np.repeat(_v, _c) for _v, _c in zip(v, c)]) for v, c in zip(vals, cnts)]
        )
        new_val, new_cnt = np.unique(comb, return_counts=True)
        new_p = new_cnt / np.sum(new_cnt)
        return new_val, new_cnt, new_p

    ## No interpolation
    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "truncnorm"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert model_distr["norm_loc"] == prop_data["norm_loc"][test_src_idx, test_tgt_idx, 0]
    assert model_distr["norm_scale"] == prop_data["norm_scale"][test_src_idx, test_tgt_idx, 0]
    assert model_distr["shared_within"] == prop_data["shared_within"][test_src_idx, test_tgt_idx, 0]
    assert model_distr["min"] == prop_data["min"][test_src_idx, test_tgt_idx, 0]
    assert model_distr["max"] == prop_data["max"][test_src_idx, test_tgt_idx, 0]

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "discrete"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert np.array_equal(model_distr["val"], prop_data["val"][test_src_idx, test_tgt_idx, 0])
    assert np.array_equal(model_distr["p"], prop_data["p"][test_src_idx, test_tgt_idx, 0])
    assert model_distr["shared_within"] == prop_data["shared_within"][test_src_idx, test_tgt_idx, 0]

    ## Level0 interpolation (source m-type & target layer/synapse class value)
    prop_data["mean"][
        test_src_idx, test_tgt_idx, 0
    ] = np.nan  # NOTE: Based on 'mean', the interpolation strategy (level) will be determined
    nsyn_data["mean"][test_src_idx, test_tgt_idx] = np.nan
    tgt_sel = np.logical_and(
        np.array(m_type_layer[1]) == m_type_layer[1][test_tgt_idx],
        np.array(m_type_class[1]) == m_type_class[1][test_tgt_idx],
    )

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "truncnorm"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert model_distr["norm_loc"] == np.nanmean(prop_data["norm_loc"][test_src_idx, tgt_sel, 0])
    assert model_distr["norm_scale"] == np.nanmean(
        prop_data["norm_scale"][test_src_idx, tgt_sel, 0]
    )
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][test_src_idx, tgt_sel, 0]).astype(bool)
    )  # Majority vote
    assert model_distr["min"] == np.nanmin(prop_data["min"][test_src_idx, tgt_sel, 0])
    assert model_distr["max"] == np.nanmax(prop_data["max"][test_src_idx, tgt_sel, 0])

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "discrete"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    val, _, p = comb_discrete(
        prop_data["val"][test_src_idx, tgt_sel, 0], prop_data["cnt"][test_src_idx, tgt_sel, 0]
    )
    assert np.array_equal(model_distr["val"], val)
    assert np.array_equal(model_distr["p"], p)
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][test_src_idx, tgt_sel, 0]).astype(bool)
    )  # Majority vote

    ## Level1 interpolation (source m-type & target synapse class value)
    prop_data["mean"][
        test_src_idx, tgt_sel, 0
    ] = np.nan  # NOTE: Based on 'mean', the interpolation strategy (level) will be determined
    nsyn_data["mean"][test_src_idx, tgt_sel] = np.nan
    tgt_sel = np.array(m_type_class[1]) == m_type_class[1][test_tgt_idx]

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "truncnorm"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert model_distr["norm_loc"] == np.nanmean(prop_data["norm_loc"][test_src_idx, tgt_sel, 0])
    assert model_distr["norm_scale"] == np.nanmean(
        prop_data["norm_scale"][test_src_idx, tgt_sel, 0]
    )
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][test_src_idx, tgt_sel, 0]).astype(bool)
    )  # Majority vote
    assert model_distr["min"] == np.nanmin(prop_data["min"][test_src_idx, tgt_sel, 0])
    assert model_distr["max"] == np.nanmax(prop_data["max"][test_src_idx, tgt_sel, 0])

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "discrete"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    val, _, p = comb_discrete(
        prop_data["val"][test_src_idx, tgt_sel, 0], prop_data["cnt"][test_src_idx, tgt_sel, 0]
    )
    assert np.array_equal(model_distr["val"], val)
    assert np.array_equal(model_distr["p"], p)
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][test_src_idx, tgt_sel, 0]).astype(bool)
    )  # Majority vote

    ## Level2 interpolation (source & target per layer/synapse class value)
    prop_data["mean"][
        test_src_idx, tgt_sel, 0
    ] = np.nan  # NOTE: Based on 'mean', the interpolation strategy (level) will be determined
    nsyn_data["mean"][test_src_idx, tgt_sel] = np.nan
    src_sel = np.logical_and(
        np.array(m_type_layer[0]) == m_type_layer[0][test_src_idx],
        np.array(m_type_class[0]) == m_type_class[0][test_src_idx],
    )
    tgt_sel = np.logical_and(
        np.array(m_type_layer[1]) == m_type_layer[1][test_tgt_idx],
        np.array(m_type_class[1]) == m_type_class[1][test_tgt_idx],
    )

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "truncnorm"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert model_distr["norm_loc"] == np.nanmean(prop_data["norm_loc"][src_sel, :, 0][:, tgt_sel])
    assert model_distr["norm_scale"] == np.nanmean(
        prop_data["norm_scale"][src_sel, :, 0][:, tgt_sel]
    )
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][src_sel, :, 0][:, tgt_sel]).astype(bool)
    )  # Majority vote
    assert model_distr["min"] == np.nanmin(prop_data["min"][src_sel, :, 0][:, tgt_sel])
    assert model_distr["max"] == np.nanmax(prop_data["max"][src_sel, :, 0][:, tgt_sel])

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "discrete"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    val, _, p = comb_discrete(
        prop_data["val"][src_sel, :, 0][:, tgt_sel], prop_data["cnt"][src_sel, :, 0][:, tgt_sel]
    )
    assert np.array_equal(model_distr["val"], val)
    assert np.array_equal(model_distr["p"], p)
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][src_sel, :, 0][:, tgt_sel]).astype(bool)
    )  # Majority vote

    ## Level3 interpolation (source & target per synapse class value)
    for tidx in np.where(tgt_sel)[0]:
        prop_data["mean"][
            src_sel, tidx, 0
        ] = np.nan  # NOTE: Based on 'mean', the interpolation strategy (level) will be determined
        nsyn_data["mean"][src_sel, tidx] = np.nan
    src_sel = np.array(m_type_class[0]) == m_type_class[0][test_src_idx]
    tgt_sel = np.array(m_type_class[1]) == m_type_class[1][test_tgt_idx]

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "truncnorm"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert model_distr["norm_loc"] == np.nanmean(prop_data["norm_loc"][src_sel, :, 0][:, tgt_sel])
    assert model_distr["norm_scale"] == np.nanmean(
        prop_data["norm_scale"][src_sel, :, 0][:, tgt_sel]
    )
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][src_sel, :, 0][:, tgt_sel]).astype(bool)
    )  # Majority vote
    assert model_distr["min"] == np.nanmin(prop_data["min"][src_sel, :, 0][:, tgt_sel])
    assert model_distr["max"] == np.nanmax(prop_data["max"][src_sel, :, 0][:, tgt_sel])

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "discrete"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    val, _, p = comb_discrete(
        prop_data["val"][src_sel, :, 0][:, tgt_sel], prop_data["cnt"][src_sel, :, 0][:, tgt_sel]
    )
    assert np.array_equal(model_distr["val"], val)
    assert np.array_equal(model_distr["p"], p)
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"][src_sel, :, 0][:, tgt_sel]).astype(bool)
    )  # Majority vote

    ## Level4 interpolation (overall value)
    for tidx in np.where(tgt_sel)[0]:
        prop_data["mean"][
            src_sel, tidx, 0
        ] = np.nan  # NOTE: Based on 'mean', the interpolation strategy (level) will be determined
        nsyn_data["mean"][src_sel, tidx] = np.nan

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "truncnorm"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    assert np.isclose(model_distr["norm_loc"], np.nanmean(prop_data["norm_loc"]))
    assert np.isclose(model_distr["norm_scale"], np.nanmean(prop_data["norm_scale"]))
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"]).astype(bool)
    )  # Majority vote
    assert model_distr["min"] == np.nanmin(prop_data["min"])
    assert model_distr["max"] == np.nanmax(prop_data["max"])

    res = test_module.build(
        nsyn_data,
        prop_data,
        m_types,
        m_type_class,
        m_type_layer,
        [prop],
        distr_types={prop: "discrete"},
        data_types={},
        data_bounds={},
    )
    model_distr = res.get_distr_props(prop_name=prop, src_type=test_src, tgt_type=test_tgt)
    val, _, p = comb_discrete(np.squeeze(prop_data["val"]), np.squeeze(prop_data["cnt"]))
    assert np.array_equal(model_distr["val"], val)
    assert np.array_equal(model_distr["p"], p)
    assert model_distr["shared_within"] == np.round(
        np.nanmean(prop_data["shared_within"]).astype(bool)
    )  # Majority vote



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import sys
import shutil
import tempfile
from contextlib import contextmanager
from types import ModuleType

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_DATA_DIR = os.path.join(TEST_DIR, "data")


@contextmanager
def setup_tempdir(prefix):
    """Create a temporary dir to be removed after use"""
    temp_dir = tempfile.mkdtemp(prefix=prefix)
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


def create_fake_module(module_name, code):
    """Used to test code that uses dynamically imported modules"""
    module = ModuleType(module_name)
    exec(code, module.__dict__)
    sys.modules[module_name] = module

    return module



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os

import morphio
import neurom as nm
import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd

from bluepysnap import Circuit
from bluepysnap.morph import MorphHelper

import pytest
import re
from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.model_building import model_types
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


# SONATA section type mapping (as in MorphIO): 1 = soma, 2 = axon, 3 = basal, 4 = apical
SEC_SOMA = 1
SEC_TYPE_MAP = {nm.AXON: 2, nm.BASAL_DENDRITE: 3, nm.APICAL_DENDRITE: 4}


@pytest.fixture
def manipulation():
    m = Manipulation.get("conn_rewiring")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)
    edges_table_empty = edges_table.loc[[]].copy()

    delay_model_file = os.path.join(
        TEST_DATA_DIR, f"model_config__DistDepDelay.json"
    )  # Deterministic delay model w/o variation
    delay_model = model_types.AbstractModel.model_from_file(delay_model_file)
    props_model_file = os.path.join(
        TEST_DATA_DIR, f"model_config__ConnProps.json"
    )  # Deterministic connection properties model w/o variation
    props_model = model_types.AbstractModel.model_from_file(props_model_file)
    props_model_file2 = os.path.join(
        TEST_DATA_DIR, f"model_config__ConnPropsNoNSynConn.json"
    )  # Deterministic connection properties model w/o variation, w/o #syn/conn
    props_model2 = model_types.AbstractModel.model_from_file(props_model_file2)

    assert not np.all(
        np.diff(edges_table["@target_node"]) >= 0
    ), "ERROR: Edges table assumed to be not sorted!"

    with pytest.raises(
        AssertionError, match=re.escape("Edges table must be ordered by @target_node!")
    ):
        with EdgeWriter(None, existing_edges=edges_table.copy()) as writer:
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class="EXC",
                prob_model_spec=None,
                delay_model_spec={"file": delay_model_file},
            )
            res = writer.to_pandas()

    # Use sorted edges table from now on
    edges_table = edges_table.sort_values(["@target_node", "@source_node"]).reset_index(drop=True)

    def check_delay(res, nodes, syn_class, delay_model):
        """Check delays (from PRE neuron (soma) to POST synapse position)"""
        for idx in res[
            np.isin(res["@source_node"], nodes[0].ids({"synapse_class": syn_class}))
        ].index:
            delay_offset = delay_model.get_param_dict()["delay_mean_coeff_a"]
            delay_scale = delay_model.get_param_dict()["delay_mean_coeff_b"]
            src_pos = nodes[0].positions(res.loc[idx]["@source_node"]).to_numpy()
            syn_pos = res.loc[idx][
                ["afferent_center_x", "afferent_center_y", "afferent_center_z"]
            ].to_numpy()
            dist = np.sqrt(np.sum((src_pos - syn_pos) ** 2))
            delay = delay_scale * dist + delay_offset
            assert np.all(np.isclose(res.loc[idx]["delay"], delay)), "ERROR: Delay mismatch!"

    def check_nsyn(ref, res):
        """Check number of synapses per connection"""
        nsyn_per_conn1 = np.unique(
            ref[["@source_node", "@target_node"]], axis=0, return_counts=True
        )[1]
        nsyn_per_conn2 = np.unique(
            res[["@source_node", "@target_node"]], axis=0, return_counts=True
        )[1]
        assert np.array_equal(
            np.sort(nsyn_per_conn1), np.sort(nsyn_per_conn2)
        ), "ERROR: Synapses per connection mismatch!"

    def check_indegree(ref, res, nodes, check_not_equal=False):
        """Check indegree"""
        indeg1 = [
            len(np.unique(ref["@source_node"][ref["@target_node"] == tid]))
            for tid in nodes[1].ids()
        ]
        indeg2 = [
            len(np.unique(res["@source_node"][res["@target_node"] == tid]))
            for tid in nodes[1].ids()
        ]
        if check_not_equal:
            assert not np.array_equal(indeg1, indeg2), "ERROR: Indegree should be different!"
        else:
            assert np.array_equal(indeg1, indeg2), "ERROR: Indegree mismatch!"

    def check_unchanged(ref, res, nodes, syn_class):
        """Check that non-target connections unchanged"""
        unch_tab1 = ref[~np.isin(ref["@source_node"], nodes[0].ids({"synapse_class": syn_class}))]
        unch_tab2 = res[~np.isin(res["@source_node"], nodes[0].ids({"synapse_class": syn_class}))]
        assert np.all(
            [
                np.sum(np.all(unch_tab1.iloc[idx] == unch_tab2, 1)) == 1
                for idx in range(unch_tab1.shape[0])
            ]
        ), f"ERROR: Non-{syn_class} connections changed!"

    def check_all_removed(ref, res, nodes, syn_class):
        """Check if all (EXC or INH) connections are removed"""
        assert (
            ref[~np.isin(ref["@source_node"], nodes[0].ids({"synapse_class": syn_class}))][
                res.columns
            ]
            .reset_index(drop=True)
            .equals(res)
        ), f"ERROR: Removed {syn_class} connections mismatch!"

    def check_sampling(ref, res, col_sel):
        """Check if synapse properties (incl. #syn/conn) sampled from existing values"""
        nsyn_per_conn1 = np.unique(
            ref[["@source_node", "@target_node"]], axis=0, return_counts=True
        )[1]
        nsyn_per_conn2 = np.unique(
            res[["@source_node", "@target_node"]], axis=0, return_counts=True
        )[1]
        assert np.all(
            np.isin(nsyn_per_conn2, nsyn_per_conn1)
        ), "ERROR: Synapses per connection sampling error!"  # Check sampling (#syn/conn)
        assert np.all(
            [np.all(np.isin(np.unique(res[col]), np.unique(ref[col]))) for col in col_sel]
        ), "ERROR: Synapse properties sampling error!"  # Check sampling (w/o #syn/conn)

    def check_zero(res, cols_unused, nodes, syn_class):
        """Check if unused properties of all rewired (EXC or INH) connections are set to zero"""
        src_sel = nodes[0].ids({"synapse_class": syn_class})
        assert np.all(
            res[np.isin(res["@source_node"], src_sel)][cols_unused] == 0
        ), "ERROR: Unused properties not set to zero!"

    def check_all_to_all(ref, res, nodes, syn_class, props_model=None, reuse_pos=True):
        """Check if all-to-all connectivity (incl. #syn/conn, if props_model provided)"""
        src_sel = nodes[0].ids({"synapse_class": syn_class})
        if reuse_pos:
            # When reusing positions, consider only target node with existing synapses; for all others, no rewiring is possible!!
            tgt_sel = np.unique(ref["@target_node"])
        else:
            # Otherwise, consider all target nodes
            tgt_sel = nodes[1].ids()
        cnt_mat = np.zeros((len(src_sel), len(tgt_sel)), dtype=int)
        if props_model is not None:
            cnt_mat_model = np.zeros((len(src_sel), len(tgt_sel)), dtype=int)
        eq_mat = np.zeros((len(src_sel), len(tgt_sel)), dtype=bool)
        for sidx, s in enumerate(src_sel):
            for tidx, t in enumerate(tgt_sel):
                cnt_mat[sidx, tidx] = np.sum(
                    np.logical_and(res["@source_node"] == s, res["@target_node"] == t)
                )
                if props_model is not None:
                    cnt_mat_model[sidx, tidx] = props_model.draw(
                        prop_name="n_syn_per_conn",
                        src_type=nodes[0].get(s, properties="mtype"),
                        tgt_type=nodes[1].get(t, properties="mtype"),
                    )[0]
                eq_mat[sidx, tidx] = s == t
        assert np.sum(cnt_mat[eq_mat]) == 0, "ERROR: Autapses found!"
        assert np.all(cnt_mat[~eq_mat] > 0), "ERROR: All-to-all connectivity expected!"
        if props_model is not None:
            assert np.all(
                cnt_mat[~eq_mat] == cnt_mat_model[~eq_mat]
            ), "ERROR: Number of synapses per connection not consistent with model!"

    def check_randomization(ref, res, props_model):
        """Check model-based randomization of synaptic properties (w/o #syn/conn)"""
        for i in range(res.shape[0]):
            r = res.iloc[i]
            if np.any(np.all(r == ref, 1)):
                continue  # Original synapse, i.e., not rewired
            src_type = nodes[0].get(int(r["@source_node"]), properties="mtype")
            tgt_type = nodes[1].get(int(r["@target_node"]), properties="mtype")
            for p in np.setdiff1d(props_model.get_prop_names(), "n_syn_per_conn"):
                assert (
                    r[p] == props_model.draw(prop_name=p, src_type=src_type, tgt_type=tgt_type)[0]
                )  # Assuming constant model!!

    def get_adj(edges_table, src_ids, tgt_ids):
        """Extract adjacency matrix from edges table"""
        conns = np.unique(edges_table[["@source_node", "@target_node"]], axis=0)
        adj_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=bool)
        for _s, _t in conns:
            adj_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = True
        return adj_mat

    def get_syn(edges_table, src_ids, tgt_ids):
        """Extract synaptome matrix from edges table"""
        conns, counts = np.unique(
            edges_table[["@source_node", "@target_node"]], axis=0, return_counts=True
        )
        syn_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=int)
        for (_s, _t), _c in zip(conns, counts):
            syn_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = _c
        return syn_mat

    def check_adj(res, edges_table, nodes, syn_class, compare_to="ref"):
        """Check adj. matrices (excl. tgt nodes w/o any original synapses; excl. autapses)"""
        _src = nodes[0].ids({"synapse_class": syn_class})
        _tgt = np.unique(edges_table["@target_node"])
        res_adj = get_adj(res, _src, _tgt)
        ref_adj = get_adj(edges_table, _src, _tgt)
        for _sidx, _s in enumerate(_src):
            for _tidx, _t in enumerate(_tgt):
                if _s == _t:  # Skip autapses
                    continue
                if compare_to == "ref":  # Compare to reference
                    assert res_adj[_sidx, _tidx] == ref_adj[_sidx, _tidx]
                elif compare_to == "inv_ref":  # Compare to inverse reference
                    assert res_adj[_sidx, _tidx] == ~ref_adj[_sidx, _tidx]
                else:  # Expect boolean value to compare to
                    assert isinstance(compare_to, bool)
                    assert res_adj[_sidx, _tidx] == compare_to

    for syn_class in ["EXC", "INH"]:
        # Case 1: Rewire connectivity with conn. prob. p=1.0 but no target selection => Nothing should be changed
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb1p0.json")
        pct = 0.0

        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=True,
            reuse_conns=True,
            gen_method=None,
            props_model_spec=None,
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert res.equals(
            edges_table[res.columns].reset_index(drop=True)
        ), "ERROR: Edges table has changed!"

        # Case 2: Rewire connectivity with conn. prob. p=0.0 (no connectivity)
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb0p0.json")
        pct = 100.0

        ## (a) Keeping indegree => NOT POSSIBLE with p=0.0
        with pytest.raises(
            AssertionError,
            match=re.escape("Keeping indegree not possible since connection probability zero!"),
        ):
            with EdgeWriter(None, existing_edges=edges_table.copy()) as writer:
                manipulation(nodes, writer).apply(
                    tgt_ids,
                    syn_class=syn_class,
                    prob_model_spec={"file": prob_model_file},
                    delay_model_spec={"file": delay_model_file},
                    sel_src=None,
                    sel_dest=None,
                    amount_pct=pct,
                    keep_indegree=True,
                    reuse_conns=True,
                    gen_method=None,
                    props_model_spec=None,
                    pos_map_file=None,
                )

        ## (b) Not keeping indegree => All selected (EXC or INH) connections should be removed
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            gen_method="sample",
            props_model_spec=None,
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_removed(edges_table, res, nodes, syn_class)

        # Case 3: Rewire connectivity with conn. prob. p=1.0 (full connectivity, w/o autapses)
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb1p0.json")
        pct = 100.0

        ## (a) Keeping indegree & reusing connections
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=True,
            reuse_conns=True,
            gen_method=None,
            props_model_spec=None,
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert np.array_equal(edges_table.shape, res.shape), "ERROR: Number of synapses mismatch!"

        col_sel = np.setdiff1d(edges_table.columns, ["@source_node", "@target_node", "delay"])
        assert edges_table[col_sel].equals(res[col_sel]), "ERROR: Synapse properties mismatch!"

        check_nsyn(edges_table, res)  # Check reuse_conns option
        check_indegree(edges_table, res, nodes)  # Check keep_indegree option
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged

        ## (b) Keeping indegree & w/o reusing connections
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=True,
            reuse_conns=False,
            gen_method="sample",
            props_model_spec=None,
            pos_map_file=None,
        )
        res = writer.to_pandas()

        # Selected columns excluding unused efferent/spine_length properties, which are not duplicated (syn_pos_mode="reuse" by default) and set to zero
        eff_props = [_col for _col in edges_table.columns if "efferent_" in _col]
        cols_unused_reuse = eff_props + ["spine_length"]
        col_sel = np.setdiff1d(
            edges_table.columns, ["@source_node", "@target_node", "delay"] + cols_unused_reuse
        )

        check_indegree(edges_table, res, nodes)  # Check keep_indegree option
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_sampling(edges_table, res, col_sel)  # Check sampling method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns

        ## (c) W/o keeping indegree & w/o reusing connections ("sample" method)
        with EdgeWriter(None, existing_edges=edges_table.copy()) as writer:
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                gen_method="sample",
                props_model_spec=None,
                pos_map_file=None,
            )
            res = writer.to_pandas()

        check_all_to_all(edges_table, res, nodes, syn_class)  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_sampling(edges_table, res, col_sel)  # Check sampling method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns

        ## (d) W/o keeping indegree & w/o reusing connections ("sample" method & block-based processing)
        split_ids_list = [tgt_ids[: len(tgt_ids) >> 1], tgt_ids[len(tgt_ids) >> 1 :]]
        res_list = []
        for i_split, split_ids in enumerate(split_ids_list):
            edges_table_split = edges_table[np.isin(edges_table["@target_node"], split_ids)].copy()
            writer = EdgeWriter(None, edges_table_split)
            manipulation(nodes, writer, i_split, len(split_ids_list)).apply(
                split_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                gen_method="sample",
                props_model_spec=None,
                pos_map_file=None,
            )
            res_list.append(writer.to_pandas())
        res = pd.concat(res_list, ignore_index=True)

        check_all_to_all(edges_table, res, nodes, syn_class)  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_sampling(edges_table, res, col_sel)  # Check sampling method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns

        ## (e) W/o keeping indegree & w/o reusing connections ("randomize" method)
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()

        check_all_to_all(
            edges_table, res, nodes, syn_class, props_model
        )  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_randomization(edges_table, res, props_model)  # Check randomization method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns

        ## (f) W/o keeping indegree & w/o reusing connections ("randomize" method & block-based processing)
        split_ids_list = [tgt_ids[: len(tgt_ids) >> 1], tgt_ids[len(tgt_ids) >> 1 :]]
        res_list = []
        for i_split, split_ids in enumerate(split_ids_list):
            edges_table_split = edges_table[np.isin(edges_table["@target_node"], split_ids)].copy()
            writer = EdgeWriter(None, edges_table_split)
            manipulation(nodes, writer, i_split, len(split_ids_list)).apply(
                split_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                gen_method="randomize",
                props_model_spec={"file": props_model_file},
                pos_map_file=None,
            )
            res_list.append(writer.to_pandas())
        res = pd.concat(res_list, ignore_index=True)

        check_all_to_all(
            edges_table, res, nodes, syn_class, props_model
        )  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_randomization(edges_table, res, props_model)  # Check randomization method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns

        # Case 4: Keeping indegree & add/delete only => NOT POSSIBLE!
        pct = 100.0
        for _mode in ["add_only", "delete_only"]:
            with pytest.raises(
                AssertionError,
                match=re.escape(f'"keep_indegree" not supported for rewire mode "{_mode}"!'),
            ):
                with EdgeWriter(None, existing_edges=edges_table.copy()) as writer:
                    manipulation(nodes, writer).apply(
                        tgt_ids,
                        syn_class=syn_class,
                        prob_model_spec={"file": prob_model_file},
                        delay_model_spec={"file": delay_model_file},
                        sel_src=None,
                        sel_dest=None,
                        amount_pct=pct,
                        keep_indegree=True,
                        rewire_mode=_mode,
                        reuse_conns=False,
                        gen_method=None,
                        props_model_spec=None,
                        pos_map_file=None,
                    )

        # Case 5: Deterministic rewiring (+ keep_conns) with empty adj. matrix
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMatEmpty.json")

        ## (a) Default rewire mode => All (EXC or INH) connections should be removed
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode=None,
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_removed(edges_table, res, nodes, syn_class)

        ## (b) Rewire mode "add_only" => All connections should be unchanged
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode="add_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert edges_table.equals(res), "ERROR: Edges table mismatch!"

        ## (c) Rewire mode "delete_only" => All (EXC or INH) connections should be removed
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode="delete_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_removed(edges_table, res, nodes, syn_class)

        # Case 6: Deterministic rewiring (+ keep_conns) with full adj. matrix
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMatFull.json")

        ## (a) Default rewire mode => All-to-all (EXC or INH) connections should exist
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode=None,
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_to_all(edges_table, res, nodes, syn_class)
        for _sclass in ["EXC", "INH"]:
            # With "keep_conns", all existing synapses should be unchanged
            check_unchanged(edges_table, res, nodes, _sclass)

        ## (b) Rewire mode "add_only" => All-to-all (EXC or INH) connections should exist
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode="add_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_to_all(edges_table, res, nodes, syn_class)
        for _sclass in ["EXC", "INH"]:
            # With "keep_conns", all existing synapses should be unchanged
            check_unchanged(edges_table, res, nodes, _sclass)

        ## (c) Rewire mode "delete_only" => All connections should be unchanged
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode="delete_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert edges_table.equals(res), "ERROR: Edges table mismatch!"

        # Case 7: Deterministic rewiring (+ keep_conns) with actual adj. matrix
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMat.json")

        ## All rewire modes => All connections should be unchanged
        for _mode in [None, "add_only", "delete_only"]:
            writer = EdgeWriter(None, edges_table.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                keep_conns=True,
                rewire_mode=_mode,
                gen_method="randomize",
                props_model_spec={"file": props_model_file},
                pos_map_file=None,
            )
            res = writer.to_pandas()
            assert edges_table.equals(res), "ERROR: Edges table mismatch!"

        # Case 8: Deterministic rewiring (+ keep_conns) with actual inverted adj. matrix
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMatInv.json")

        ## (a) Default rewire mode => Connectivity should be inverted (EXC or INH)
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode=None,
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_adj(res, edges_table, nodes, syn_class, compare_to="inv_ref")

        ## (b) Rewire mode "add_only" => All-to-all (EXC or INH) connections should exist
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode="add_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_to_all(edges_table, res, nodes, syn_class)
        for _sclass in ["EXC", "INH"]:
            # With "keep_conns", all existing synapses should be unchanged
            check_unchanged(edges_table, res, nodes, _sclass)

        ## (c) Rewire mode "delete_only" => All (EXC or INH) connections should be removed
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=True,
            rewire_mode="delete_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_removed(edges_table, res, nodes, syn_class)

        # Case 9: Deterministic rewiring (w/o keep/reuse_conns) with full adj. matrix
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMatFull.json")

        ## (a) Default rewire mode => All-to-all (EXC or INH) connections should exist
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            rewire_mode=None,
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_to_all(edges_table, res, nodes, syn_class)
        check_adj(res, edges_table, nodes, syn_class, compare_to=True)

        ## (b) Rewire mode "add_only" => All-to-all (EXC or INH) connections should exist
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            rewire_mode="add_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        check_all_to_all(edges_table, res, nodes, syn_class)
        check_adj(res, edges_table, nodes, syn_class, compare_to=True)

        ## (c) Rewire mode "delete_only" => Connectivity should be unchanged, but connections re-drawn
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            rewire_mode="delete_only",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert not edges_table.equals(res), "ERROR: Edges tables should not be equal!"
        check_adj(res, edges_table, nodes, syn_class, compare_to="ref")

        # Case 10: Deterministic rewiring (w/o keep/reuse_conns) with actual adj. matrix
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMat.json")

        ## All rewire modes => Connectivity should be unchanged, but connections re-drawn
        for _mode in [None, "add_only", "delete_only"]:
            writer = EdgeWriter(None, edges_table.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                keep_conns=False,
                rewire_mode=_mode,
                gen_method="randomize",
                props_model_spec={"file": props_model_file},
                pos_map_file=None,
            )
            res = writer.to_pandas()
            assert not edges_table.equals(res), "ERROR: Edges tables should not be equal!"
            check_adj(res, edges_table, nodes, syn_class, compare_to="ref")

        # Case 11: Synapse placement using morphologies (w/o reusing positions)
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb1p0.json")
        pct = 100.0

        tgt_morph = MorphHelper(
            nodes[1].config.get("morphologies_dir"),
            nodes[1],
            nodes[1].config.get("alternate_morphologies"),
        )
        get_tgt_morph = lambda node_id: tgt_morph.get(
            node_id, transform=True, extension="swc"
        )  # Access function (incl. transformation!), using specified format (swc/h5/...)

        morphio.set_maximum_warnings(1)  # Suppress repeated "Warning: zero diameter in file"

        eff_props = [_col for _col in edges_table.columns if "efferent_" in _col]
        seg_props = [_col for _col in edges_table.columns if "_segment" in _col]
        surf_props = [_col for _col in edges_table.columns if "_surface" in _col]
        cols_unused_all = (
            eff_props + seg_props + surf_props + ["spine_length"]
        )  # W/o reusing positions, more columns will be unused
        aff_props = [_col for _col in edges_table.columns if "afferent_" in _col]
        col_sel = np.setdiff1d(
            edges_table.columns,
            ["@source_node", "@target_node", "delay"] + cols_unused_all + aff_props,
        )

        def check_syn_pos(res, nodes, syn_class):
            """Check synapse position/type consistency"""
            for i in range(res.shape[0]):
                syn_cl = (
                    nodes[0]
                    .get(res.iloc[i]["@source_node"], properties="synapse_class")
                    .to_numpy()[0]
                )
                if syn_cl != syn_class:
                    continue

                # Check synapse position/type consistency
                syn_pos = res.iloc[i][
                    ["afferent_center_x", "afferent_center_y", "afferent_center_z"]
                ]
                sec_id, sec_pos, sec_type = res.iloc[i][
                    ["afferent_section_id", "afferent_section_pos", "afferent_section_type"]
                ]
                if sec_id == 0:  # Soma section
                    assert sec_pos == 0.0 and sec_type == SEC_SOMA, "ERROR: Soma section error!"
                    assert np.all(
                        np.isclose(
                            syn_pos.to_numpy(),
                            nodes[1].positions(res.iloc[i]["@target_node"]).to_numpy(),
                        )
                    ), "ERROR: Soma position error!"
                else:
                    morph = get_tgt_morph(int(res.iloc[i]["@target_node"]))
                    sec_id = int(
                        sec_id - 1
                    )  # IMPORTANT: Section IDs in NeuroM morphology don't include soma, so they need to be shifted by 1 (Soma ID is 0 in edges table)
                    assert (
                        sec_type == SEC_TYPE_MAP[morph.section(sec_id).type]
                    ), "ERROR: Section type mismatch!"
                    assert np.all(
                        np.isclose(
                            nm.morphmath.path_fraction_point(morph.section(sec_id).points, sec_pos),
                            syn_pos,
                        )
                    ), "ERROR: Section position error!"

        ## (a) W/o keeping indegree & w/o reusing connections ("sample" method)
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            syn_pos_mode="random",
            gen_method="sample",
            props_model_spec=None,
            pos_map_file=None,
        )
        res = writer.to_pandas()

        check_all_to_all(
            edges_table, res, nodes, syn_class, reuse_pos=False
        )  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_sampling(edges_table, res, col_sel)  # Check sampling method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        check_syn_pos(res, nodes, syn_class)  # Check synapse positions

        ## (b) W/o keeping indegree & w/o reusing connections ("sample" method & block-based processing)
        split_ids_list = [tgt_ids[: len(tgt_ids) >> 1], tgt_ids[len(tgt_ids) >> 1 :]]
        res_list = []
        for i_split, split_ids in enumerate(split_ids_list):
            edges_table_split = edges_table[np.isin(edges_table["@target_node"], split_ids)].copy()
            writer = EdgeWriter(None, edges_table_split)
            manipulation(nodes, writer, i_split, len(split_ids_list)).apply(
                split_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                syn_pos_mode="random",
                gen_method="sample",
                props_model_spec=None,
                pos_map_file=None,
            )
            res_list.append(writer.to_pandas())
        res = pd.concat(res_list, ignore_index=True)

        check_all_to_all(
            edges_table, res, nodes, syn_class, reuse_pos=False
        )  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_sampling(edges_table, res, col_sel)  # Check sampling method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        check_syn_pos(res, nodes, syn_class)  # Check synapse positions

        ## (c) W/o keeping indegree & w/o reusing connections ("randomize" method)
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            syn_pos_mode="random",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()

        check_all_to_all(
            edges_table, res, nodes, syn_class, props_model, reuse_pos=False
        )  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_randomization(edges_table, res, props_model)  # Check randomization method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        check_syn_pos(res, nodes, syn_class)  # Check synapse positions

        ## (d) W/o keeping indegree & w/o reusing connections ("randomize" method & block-based processing)
        split_ids_list = [tgt_ids[: len(tgt_ids) >> 1], tgt_ids[len(tgt_ids) >> 1 :]]
        res_list = []
        for i_split, split_ids in enumerate(split_ids_list):
            edges_table_split = edges_table[np.isin(edges_table["@target_node"], split_ids)].copy()
            writer = EdgeWriter(None, edges_table_split)
            manipulation(nodes, writer, i_split, len(split_ids_list)).apply(
                split_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                syn_pos_mode="random",
                gen_method="randomize",
                props_model_spec={"file": props_model_file},
                pos_map_file=None,
            )
            res_list.append(writer.to_pandas())
        res = pd.concat(res_list, ignore_index=True)

        check_all_to_all(
            edges_table, res, nodes, syn_class, props_model, reuse_pos=False
        )  # Check all-to-all connectivity
        check_indegree(
            edges_table, res, nodes, check_not_equal=True
        )  # Check if keep_indegree changed
        check_unchanged(
            edges_table, res, nodes, syn_class
        )  # Check that non-selected connections unchanged
        check_randomization(edges_table, res, props_model)  # Check randomization method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        check_syn_pos(res, nodes, syn_class)  # Check synapse positions

        # Case 12: Wiring an empty connectome from scratch
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__ConnProb1p0.json")
        pct = 100.0

        ## (a) Reusing positions => Not possible, so output must be empty
        writer = EdgeWriter(None, edges_table_empty.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            syn_pos_mode="reuse",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert edges_table_empty.equals(res), "ERROR: Edges not empty!"

        ## (b) Keeping indegree => Nothing to wire, so output must be empty
        writer = EdgeWriter(None, edges_table_empty.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=True,
            reuse_conns=False,
            syn_pos_mode="random",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()
        assert edges_table_empty.equals(res), "ERROR: Edges not empty!"

        ## (c) Sampling method => ERROR since no synapses to sample values from
        with pytest.raises(
            AssertionError,
            match=re.escape(
                f"No synapses to sample connection property values for target neuron {tgt_ids[0]} from!"
            ),
        ):
            writer = EdgeWriter(None, edges_table_empty.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                sel_src=None,
                sel_dest=None,
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                syn_pos_mode="random",
                gen_method="sample",
                props_model_spec=None,
                pos_map_file=None,
            )

        ## (d) Wiring from scratch => All-to-all connectivity
        writer = EdgeWriter(None, edges_table_empty.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            sel_src=None,
            sel_dest=None,
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            syn_pos_mode="random",
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            pos_map_file=None,
        )
        res = writer.to_pandas()

        check_all_to_all(
            edges_table, res, nodes, syn_class, props_model, reuse_pos=False
        )  # Check all-to-all connectivity
        check_randomization(edges_table, res, props_model)  # Check randomization method
        check_delay(res, nodes, syn_class, delay_model)  # Check synaptic delays
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        check_syn_pos(res, nodes, syn_class)  # Check synapse positions

        # Case 14: Deterministic rewiring with actual adj. matrix and synaptome provided
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMat.json")
        syn_model_file = os.path.join(TEST_DATA_DIR, "model_config__SynMatRnd.json")
        syn_model = model_types.AbstractModel.init_model({"file": syn_model_file})

        ## (a) Provide synaptome model, #syn/conn from ConnPropsModel ignored
        #      => Adjacency should be unchanged, but #syn/conn from synaptome
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            gen_method="randomize",
            props_model_spec={"file": props_model_file},
            nsynconn_model_spec={"file": syn_model_file},
        )
        res = writer.to_pandas()
        assert not edges_table.equals(res), "ERROR: Edges tables should not be equal!"
        check_adj(res, edges_table, nodes, syn_class, compare_to="ref")
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns
        src_ids_sel = nodes[0].ids({"synapse_class": syn_class})
        res_syn = get_syn(res, src_ids_sel, tgt_ids)  # synaptome
        assert_array_equal(res_syn, syn_model.apply(src_nid=src_ids_sel, tgt_nid=tgt_ids))

        ## (b) Provide synaptome model & ConnPropsModel w/o #syn/conn
        #      => Adjacency should be unchanged, but #syn/conn from synaptome
        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            gen_method="randomize",
            props_model_spec={"file": props_model_file2},
            nsynconn_model_spec={"file": syn_model_file},
        )
        res = writer.to_pandas()
        assert not edges_table.equals(res), "ERROR: Edges tables should not be equal!"
        check_adj(res, edges_table, nodes, syn_class, compare_to="ref")
        check_zero(res, cols_unused_reuse, nodes, syn_class)  # Check unused columns
        src_ids_sel = nodes[0].ids({"synapse_class": syn_class})
        res_syn = get_syn(res, src_ids_sel, tgt_ids)  # synaptome
        assert_array_equal(res_syn, syn_model.apply(src_nid=src_ids_sel, tgt_nid=tgt_ids))

        ## (c) Provide ConnPropsModel w/o #syn/conn, but no synaptome => Assertion error
        with pytest.raises(
            AssertionError,
            match=re.escape("#Syn/conn model required when using a properties model w/o nsynconn!"),
        ):
            writer = EdgeWriter(None, edges_table.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                keep_conns=False,
                gen_method="randomize",
                props_model_spec={"file": props_model_file2},
            )

        ## (d) Provide empty synaptome model => Assertion error
        with pytest.raises(AssertionError, match=re.escape('"n_syn" must be at least 1!')):
            syn_model_file = os.path.join(TEST_DATA_DIR, "model_config__SynMatEmpty.json")
            writer = EdgeWriter(None, edges_table.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                keep_conns=False,
                gen_method="randomize",
                props_model_spec={"file": props_model_file},
                nsynconn_model_spec={"file": syn_model_file},
            )

        # Case 15: Deterministic rewiring with actual adj. matrix and new synapse positions externally provided
        pct = 100.0
        prob_model_file = os.path.join(TEST_DATA_DIR, "model_config__AdjMat.json")
        syn_pos_file = os.path.join(TEST_DATA_DIR, "model_config__SynPosTable.json")
        syn_pos = model_types.AbstractModel.init_model({"file": syn_pos_file})

        def check_pos(syn_pos, res, nodes, syn_class, ref_count=None):
            """Check synapse positions loaded from external table."""
            src_ids_sel = nodes[0].ids({"synapse_class": syn_class})
            res_conns, res_counts = np.unique(
                res.loc[
                    np.isin(res["@source_node"], src_ids_sel), ["@source_node", "@target_node"]
                ],
                axis=0,
                return_counts=True,
            )
            if ref_count is not None:
                assert np.all(res_counts == ref_count), "ERROR: Ref synapse count mismatch!"
            aff_props = [f"afferent_section_{_p}" for _p in ["id", "pos", "type"]] + [
                f"afferent_center_{_p}" for _p in ["x", "y", "z"]
            ]
            ref_pos = pd.concat(
                [
                    syn_pos.apply(
                        src_nid=_conn[0], tgt_nid=_conn[1], num_sel=ref_count, prop_names=aff_props
                    )
                    for _conn in res_conns
                ]
            )
            res_pos = pd.concat(
                [
                    res.loc[
                        np.logical_and(
                            res["@source_node"] == _conn[0], res["@target_node"] == _conn[1]
                        ),
                        aff_props,
                    ]
                    for _conn in res_conns
                ]
            )
            assert_array_equal(res_pos.to_numpy(), ref_pos.to_numpy())

        # (a) Generate 10 synapses per connection => Assertion error since not enough positions provided
        syn_model_file = os.path.join(TEST_DATA_DIR, "model_config__SynMatTen.json")

        with pytest.raises(ValueError):
            writer = EdgeWriter(None, edges_table.copy())
            manipulation(nodes, writer).apply(
                tgt_ids,
                syn_class=syn_class,
                prob_model_spec={"file": prob_model_file},
                delay_model_spec={"file": delay_model_file},
                amount_pct=pct,
                keep_indegree=False,
                reuse_conns=False,
                keep_conns=False,
                syn_pos_mode="external",
                gen_method="randomize",
                props_model_spec={"file": props_model_file2},
                nsynconn_model_spec={"file": syn_model_file},
                syn_pos_model_spec={"file": syn_pos_file},
            )

        # (b) Only generate 1 synapse per connection => First pos from table must be loaded
        syn_model_file = os.path.join(TEST_DATA_DIR, "model_config__SynMatOne.json")
        syn_model = model_types.AbstractModel.init_model({"file": syn_model_file})

        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            syn_pos_mode="external",
            gen_method="randomize",
            props_model_spec={"file": props_model_file2},
            nsynconn_model_spec={"file": syn_model_file},
            syn_pos_model_spec={"file": syn_pos_file},
        )
        res = writer.to_pandas()
        assert not edges_table.equals(res), "ERROR: Edges tables should not be equal!"
        check_adj(res, edges_table, nodes, syn_class, compare_to="ref")
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        src_ids_sel = nodes[0].ids({"synapse_class": syn_class})
        res_syn = get_syn(res, src_ids_sel, tgt_ids)  # synaptome
        assert res_syn.max() == 1
        assert_array_equal(res_syn, syn_model.apply(src_nid=src_ids_sel, tgt_nid=tgt_ids))
        check_pos(syn_pos, res, nodes, syn_class, ref_count=1)

        # (c) Use exact synaptome => Full pos table must be loaded
        syn_model_file = os.path.join(TEST_DATA_DIR, "model_config__SynMat.json")
        syn_model = model_types.AbstractModel.init_model({"file": syn_model_file})

        writer = EdgeWriter(None, edges_table.copy())
        manipulation(nodes, writer).apply(
            tgt_ids,
            syn_class=syn_class,
            prob_model_spec={"file": prob_model_file},
            delay_model_spec={"file": delay_model_file},
            amount_pct=pct,
            keep_indegree=False,
            reuse_conns=False,
            keep_conns=False,
            syn_pos_mode="external",
            gen_method="randomize",
            props_model_spec={"file": props_model_file2},
            nsynconn_model_spec={"file": syn_model_file},
            syn_pos_model_spec={"file": syn_pos_file},
        )
        res = writer.to_pandas()
        assert not edges_table.equals(res), "ERROR: Edges tables should not be equal!"
        check_adj(res, edges_table, nodes, syn_class, compare_to="ref")
        check_zero(res, cols_unused_all, nodes, syn_class)  # Check unused columns
        src_ids_sel = nodes[0].ids({"synapse_class": syn_class})
        res_syn = get_syn(res, src_ids_sel, tgt_ids)  # synaptome
        assert_array_equal(res_syn, syn_model.apply(src_nid=src_ids_sel, tgt_nid=tgt_ids))
        check_pos(syn_pos, res, nodes, syn_class)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os
import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal
from bluepysnap import Circuit
import pytest
from utils import setup_tempdir, TEST_DATA_DIR

import connectome_manipulator.model_building.pos_mapping_from_table as test_module


def test_extract():
    np.random.seed(0)
    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = c.nodes["nodeA"]
    node_ids = nodes.ids()
    with setup_tempdir(__name__) as tempdir:
        N = 100
        coord_names = ["xpos", "ypos", "zpos"]

        # Check basic extraction (zero-based indexing)
        pos_tab = pd.DataFrame(np.random.rand(N, len(coord_names)), columns=coord_names)
        pos_file = os.path.join(tempdir, "pos_tab_TEST.feather")
        pos_tab.to_feather(pos_file)

        res = test_module.extract(
            c,
            pos_file,
            coord_names,
            coord_scale=None,
            nodes_pop_name=None,
            nodes_spec=None,
            zero_based_indexing=True,
            gid_column=None,
        )
        assert_array_equal(res["nrn_ids"], node_ids)
        assert_array_equal(res["map_pos"], pos_tab[coord_names].loc[node_ids])
        assert_array_equal(res["nrn_pos"], nodes.positions(node_ids))
        assert_array_equal(res["nrn_lay"], nodes.get(node_ids, properties="layer"))

        # Check coordinate scaling (zero-based indexing)
        coord_scale = np.random.rand(len(coord_names))

        res = test_module.extract(
            c,
            pos_file,
            coord_names,
            coord_scale=coord_scale,
            nodes_pop_name=None,
            nodes_spec=None,
            zero_based_indexing=True,
            gid_column=None,
        )
        assert_array_equal(res["nrn_ids"], node_ids)
        assert_array_equal(res["map_pos"], pos_tab[coord_names].loc[node_ids] * coord_scale)
        assert_array_equal(res["nrn_pos"], nodes.positions(node_ids))
        assert_array_equal(res["nrn_lay"], nodes.get(node_ids, properties="layer"))

        # Check extraction using GID column (zero-based indexing)
        pos_tab = pd.DataFrame(
            np.hstack([np.random.rand(N, len(coord_names)), np.arange(N).reshape([N, 1])]),
            columns=coord_names + ["gid"],
        )
        pos_tab = pos_tab.astype({"gid": int})
        pos_file = os.path.join(tempdir, "pos_tab_TEST.feather")
        pos_tab.to_feather(pos_file)

        res = test_module.extract(
            c,
            pos_file,
            coord_names,
            coord_scale=None,
            nodes_pop_name=None,
            nodes_spec=None,
            zero_based_indexing=True,
            gid_column="gid",
        )
        assert_array_equal(res["nrn_ids"], node_ids)
        assert_array_equal(res["map_pos"], pos_tab[coord_names].loc[node_ids])
        assert_array_equal(res["nrn_pos"], nodes.positions(node_ids))
        assert_array_equal(res["nrn_lay"], nodes.get(node_ids, properties="layer"))

        # Check extraction using GID column (one-based indexing)
        pos_tab = pd.DataFrame(
            np.hstack([np.random.rand(N, len(coord_names)), np.arange(1, N + 1).reshape([N, 1])]),
            columns=coord_names + ["gid"],
        )
        pos_tab = pos_tab.astype({"gid": int})
        pos_file = os.path.join(tempdir, "pos_tab_TEST.feather")
        pos_tab.to_feather(pos_file)

        res = test_module.extract(
            c,
            pos_file,
            coord_names,
            coord_scale=None,
            nodes_pop_name=None,
            nodes_spec=None,
            zero_based_indexing=False,
            gid_column="gid",
        )
        assert_array_equal(res["nrn_ids"], node_ids)
        assert_array_equal(res["map_pos"], pos_tab[coord_names].loc[node_ids])
        assert_array_equal(res["nrn_pos"], nodes.positions(node_ids))
        assert_array_equal(res["nrn_lay"], nodes.get(node_ids, properties="layer"))


def test_build():
    N = 100
    coord_names = ["xpos", "ypos", "zpos"]
    nrn_ids = np.arange(N) + 5
    map_pos = np.random.rand(N, len(coord_names))

    # Check pos mapping model
    model = test_module.build(nrn_ids, coord_names, map_pos, model_coord_names=None)
    assert_array_equal(model.get_coord_names(), coord_names)
    assert_array_equal(model.get_gids(), nrn_ids)
    assert_array_equal(model.apply(gids=nrn_ids), map_pos)

    # Check model coordinate names
    model_coord_names = ["a", "b", "c"]
    model = test_module.build(nrn_ids, coord_names, map_pos, model_coord_names=model_coord_names)
    assert_array_equal(model.get_coord_names(), model_coord_names)
    assert_array_equal(model.get_gids(), nrn_ids)
    assert_array_equal(model.apply(gids=nrn_ids), map_pos)

    # Check error handling (NaN)
    map_pos = np.random.rand(N, len(coord_names))
    map_pos[np.random.choice(map_pos.shape[0]), np.random.choice(map_pos.shape[1])] = np.nan
    with pytest.raises(AssertionError, match="ERROR: Invalid mapped positions found!"):
        model = test_module.build(nrn_ids, coord_names, map_pos, model_coord_names=None)

    # Check error handling (Inf)
    map_pos = np.random.rand(N, len(coord_names))
    map_pos[np.random.choice(map_pos.shape[0]), np.random.choice(map_pos.shape[1])] = np.inf
    with pytest.raises(AssertionError, match="ERROR: Invalid mapped positions found!"):
        model = test_module.build(nrn_ids, coord_names, map_pos, model_coord_names=None)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import re
import json
import pytest
from pathlib import Path
from cached_property import cached_property
from connectome_manipulator import utils as test_module
from bluepysnap.circuit import Circuit

from utils import TEST_DATA_DIR


@pytest.mark.parametrize(
    "base_dir, path, expected",
    [
        ("/a/b", "$c", "$c"),
        ("/a/b", "", ""),
        ("/a/b", "/c/d", "/c/d"),
        ("/a/b", "/a/b/c", "$BASE_DIR/c"),
        ("/a/b", "/a/b/c/d", "$BASE_DIR/c/d"),
        ("/a/b/c", "/a/b/c/d", "$BASE_DIR/d"),
        ("/a/b", "./c", "$BASE_DIR/c"),
        ("/a/b", "c/d", "$BASE_DIR/c/d"),
    ],
)
def tested_reduce_path(base_dir, path, expected):
    result = test_module._reduce_path(path, Path(base_dir))
    assert result == expected


def test_reduce_config_paths__raises_relative_config_dir():
    expected = re.escape("Circuit config's directory is not absolute: .")
    with pytest.raises(ValueError, match=expected):
        test_module.reduce_config_paths(None, ".")


def test_reduce_config_paths__not_resolved_config():
    config_path = Path(TEST_DATA_DIR, "circuit_sonata.json")
    config = json.loads(config_path.read_bytes())
    contains = "A reduced config with absolute paths and no manifest must be provided."
    with pytest.raises(ValueError, match=contains):
        test_module.reduce_config_paths(config, "/")


def test_reduce_config_paths():
    config_path = Path(TEST_DATA_DIR, "circuit_sonata.json")

    reduced_config = Circuit(config_path).config

    res = test_module.reduce_config_paths(reduced_config, TEST_DATA_DIR)

    assert res["version"] == 2
    assert res["node_sets_file"] == "$BASE_DIR/node_sets.json"
    assert res["networks"] == {
        "nodes": [
            {
                "nodes_file": "$BASE_DIR/nodes.h5",
                "populations": {
                    "nodeA": {
                        "type": "biophysical",
                        "morphologies_dir": "$BASE_DIR/swc",
                        "biophysical_neuron_models_dir": "$BASE_DIR",
                    },
                },
            },
        ],
        "edges": [
            {
                "edges_file": "$BASE_DIR/edges.h5",
                "populations": {"nodeA__nodeA__chemical": {"type": "chemical"}},
            },
        ],
    }



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for building connection/synapse properties models"""

import os.path

import matplotlib.pyplot as plt
import numpy as np
import progressbar
from scipy.optimize import curve_fit
from scipy.stats import norm

from connectome_manipulator import log
from connectome_manipulator.model_building import model_types
from connectome_manipulator.access_functions import get_edges_population, get_node_ids, get_cv_data

MAX_UNIQUE_COUNT = 100  # To be used in discrete distributions

# Ideas for improvement:
#   *Detect actual distributions of synaptic properties (incl. data type!)
#   *Capture cross-correlations between synaptic properties


def extract(
    circuit,
    min_sample_size_per_group=None,
    max_sample_size_per_group=None,
    hist_bins=51,  # [In case of constant distributions, better to use odd bin count so that center bin symmetrically centered around constant value; otherwise, rounding problems at bin edges possible]
    sel_props=None,
    sel_src=None,
    sel_dest=None,
    edges_popul_name=None,
    CV_dict=None,
    **_,
):
    """Extracts statistics (like mean, std, min, max, histogram, ...) for synapse properties of samples of connections between each pair of m-types.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        min_sample_size_per_group (int): Minimum number of samples (connections) required based on which to estimate statistics for a given pair of m-types; otherwise, no statistics will be estimated
        max_sample_size_per_group (int): Maximum number of samples (connections) based on which to estimate statistics for a given pair of m-types; if more samples are available, a random subset is used
        hist_bins (int): Number of bins for extracting histograms
        sel_props (list-like): List of synaptic property names to extract statistics from; None to select default properties
        sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
        sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
        edges_popul_name (str): Name of SONATA egdes population to extract data from
        CV_dict (dict): Optional cross-validation dictionary, containing "n_folds" (int), "fold_idx" (int), "training_set" (bool) keys; will be automatically provided by the framework if "CV_folds" are specified

    Returns:
        dict: Dictionary containing the extracted data elements

    Note:
        The following statistics will be extracted:

        * "mean" (float): Data mean
        * "std" (float): Standard deviation
        * "min" (float): Minimum value
        * "max" (float): Maximum value
        * "hist" (tuple): Histogram with ``hist_bins`` bins; stored as (counts, edges) tuple as returned by :func:`numpy.histogram`
        * "norm_loc" (float): Location estimate for truncnorm distributions
        * "norm_scale" (float): Scale estimate for truncnorm distributions
        * "val" (list-like): List of unique values for discrete distributions (*)
        * "cnt" (list-like): List of unique value counts for discrete distributions (*)
        * "p" (list-like): List of unique value probabilities for discrete distributions (*)
        * "shared_within" (bool): Flag indicating that all synapses within a connection share the same value for a given property; not applicable for #synapses/connection

        (*) Discrete statistics (i.e., "val"/"cnt"/"p") are only stored if there are not too many (less or equal ``MAX_UNIQUE_COUNT=100``).

    Note:
        Statistics for #synapses/connection will always be extracted, even if no properties are selected under ``sel_props`` (i.e., empty list).
    """
    # Select edge population
    edges = get_edges_population(circuit, edges_popul_name)

    # Select corresponding source/target nodes populations
    src_nodes = edges.source
    tgt_nodes = edges.target
    nodes = [src_nodes, tgt_nodes]

    node_ids_src = get_node_ids(src_nodes, sel_src)
    node_ids_dest = get_node_ids(tgt_nodes, sel_dest)

    # Cross-validation (optional)
    node_ids_src, node_ids_dest = get_cv_data([node_ids_src, node_ids_dest], CV_dict)

    node_ids = [node_ids_src, node_ids_dest]

    m_types = [
        np.unique(n.get(ids, properties="mtype")).tolist() for n, ids in zip(nodes, node_ids)
    ]
    m_type_class = [
        [
            nodes[i]
            .get(
                np.intersect1d(nodes[i].ids({"mtype": m}), node_ids[i]), properties="synapse_class"
            )
            .iloc[0]
            for m in m_types[i]
        ]
        for i in range(len(nodes))
    ]
    m_type_layer = [
        [
            nodes[i]
            .get(np.intersect1d(nodes[i].ids({"mtype": m}), node_ids[i]), properties="layer")
            .iloc[0]
            for m in m_types[i]
        ]
        for i in range(len(nodes))
    ]
    if sel_props is None:  # Select all usable properties
        syn_props = list(
            filter(
                lambda x: not np.any(
                    [excl in x for excl in ["@", "delay", "afferent", "efferent", "spine_length"]]
                ),
                edges.property_names,
            )
        )
    else:
        syn_props = list(sel_props)
        syn_props_check = np.array([prop in edges.property_names for prop in syn_props])
        if len(syn_props_check) > 0:
            log.log_assert(
                np.all(syn_props_check),
                f"Selected synapse properties not found: {np.array(syn_props)[~syn_props_check]}",
            )

    log.debug(
        f"Estimating statistics for {len(syn_props)} properties (plus #synapses/connection) between {len(m_types[0])}x{len(m_types[1])} m-types (min_sample_size_per_group={min_sample_size_per_group}, max_sample_size_per_group={max_sample_size_per_group}, CV_dict={CV_dict})"
    )

    # Statistics for #syn/conn
    # (incl. fitted norm loc/scale for truncnorm distribution,
    #  incl. unique values/counts/probabilities for discrete distribution)
    syns_per_conn_data = {
        "mean": np.full((len(m_types[0]), len(m_types[1])), np.nan),
        "std": np.full((len(m_types[0]), len(m_types[1])), np.nan),
        "min": np.full((len(m_types[0]), len(m_types[1])), np.nan),
        "max": np.full((len(m_types[0]), len(m_types[1])), np.nan),
        "hist": np.full((len(m_types[0]), len(m_types[1])), np.nan, dtype=object),
        "norm_loc": np.full((len(m_types[0]), len(m_types[1])), np.nan),
        "norm_scale": np.full((len(m_types[0]), len(m_types[1])), np.nan),
        "val": np.full((len(m_types[0]), len(m_types[1])), np.nan, dtype=object),
        "cnt": np.full((len(m_types[0]), len(m_types[1])), np.nan, dtype=object),
        "p": np.full((len(m_types[0]), len(m_types[1])), np.nan, dtype=object),
    }

    # Statistics for synapse/connection properties
    # (incl. fitted norm loc/scale for truncnorm distribution,
    #  incl. unique values/counts/probabilities for discrete distribution;
    #  incl. shared_within to indicate that all synapses within a connection share same value)
    conn_prop_data = {
        "mean": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
        "std": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
        "min": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
        "max": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
        "hist": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan, dtype=object),
        "norm_loc": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
        "norm_scale": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
        "val": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan, dtype=object),
        "cnt": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan, dtype=object),
        "p": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan, dtype=object),
        "shared_within": np.full((len(m_types[0]), len(m_types[1]), len(syn_props)), np.nan),
    }

    # Extract statistics
    conn_counts = {"min": np.inf, "max": -np.inf, "sel": 0}  # Count connections for reporting
    pbar = progressbar.ProgressBar()
    for sidx in pbar(range(len(m_types[0]))):
        sids = np.intersect1d(src_nodes.ids({"mtype": m_types[0][sidx]}), node_ids_src)
        for tidx in range(len(m_types[1])):
            tids = np.intersect1d(tgt_nodes.ids({"mtype": m_types[1][tidx]}), node_ids_dest)
            edges_sel = edges.pathway_edges(
                sids, tids, ["@source_node", "@target_node"] + syn_props
            )
            if edges_sel.shape[0] == 0:  # No synapses between pair of m-types
                continue

            _, syn_conn_idx, num_syn_per_conn = np.unique(
                edges_sel[["@source_node", "@target_node"]],
                axis=0,
                return_inverse=True,
                return_counts=True,
            )
            conn_counts["min"] = min(conn_counts["min"], len(num_syn_per_conn))
            conn_counts["max"] = max(conn_counts["max"], len(num_syn_per_conn))
            conn_sel = range(len(num_syn_per_conn))  # Select all connections
            if (
                min_sample_size_per_group is not None
                and min_sample_size_per_group > 0
                and len(conn_sel) < min_sample_size_per_group
            ):
                # Not enough connections available
                continue

            if max_sample_size_per_group is not None and 0 < max_sample_size_per_group < len(
                conn_sel
            ):
                # Subsample connections
                conn_sel = sorted(
                    np.random.choice(conn_sel, max_sample_size_per_group, replace=False)
                )
            conn_counts["sel"] += 1

            syns_per_conn_data["mean"][sidx, tidx] = np.mean(num_syn_per_conn[conn_sel])
            syns_per_conn_data["std"][sidx, tidx] = np.std(num_syn_per_conn[conn_sel])
            syns_per_conn_data["min"][sidx, tidx] = np.min(num_syn_per_conn[conn_sel])
            syns_per_conn_data["max"][sidx, tidx] = np.max(num_syn_per_conn[conn_sel])
            hist_counts, bin_edges = np.histogram(num_syn_per_conn[conn_sel], bins=hist_bins)
            syns_per_conn_data["hist"][sidx, tidx] = (hist_counts, bin_edges)
            bin_centers = [np.mean(bin_edges[i : i + 2]) for i in range(len(hist_counts))]
            loc, scale, _ = _norm_fitting(
                bin_centers,
                hist_counts,
                def_mn=syns_per_conn_data["mean"][sidx, tidx],
                def_sd=syns_per_conn_data["std"][sidx, tidx],
            )
            syns_per_conn_data["norm_loc"][sidx, tidx] = loc
            syns_per_conn_data["norm_scale"][sidx, tidx] = scale
            v, c = np.unique(num_syn_per_conn[conn_sel], return_counts=True)
            if len(v) == 1 or (
                len(v) < len(num_syn_per_conn[conn_sel]) and len(v) <= MAX_UNIQUE_COUNT
            ):  # Store discrete values only if not too many [otherwise, it is probably not a discrete distribution]
                syns_per_conn_data["val"][sidx, tidx] = v
                syns_per_conn_data["cnt"][sidx, tidx] = c
                syns_per_conn_data["p"][sidx, tidx] = c / np.sum(c)
            else:
                syns_per_conn_data["val"][sidx, tidx] = []
                syns_per_conn_data["cnt"][sidx, tidx] = []
                syns_per_conn_data["p"][sidx, tidx] = []
                # log.warning(f'Discrete #synapses/connection values not stored for {m_types[0][sidx]}-{m_types[1][tidx]} ({len(v)} of {len(num_syn_per_conn[conn_sel])} unique values)!')

            # Collect property statistics over connections (if shared values) or synapses (if non-shared values)
            for pidx, p in enumerate(syn_props):
                is_shared = _check_shared_property(edges_sel, p, conn_sel, syn_conn_idx)
                prop_values = _get_property_values(edges_sel, p, conn_sel, syn_conn_idx, is_shared)

                conn_prop_data["shared_within"][sidx, tidx, pidx] = is_shared
                conn_prop_data["mean"][sidx, tidx, pidx] = np.mean(
                    prop_values, dtype=np.float64
                )  # [float64 required, otherwise rounding problems!!!]
                conn_prop_data["std"][sidx, tidx, pidx] = np.std(
                    prop_values, dtype=np.float64
                )  # [float64 required, otherwise rounding problems!!!]
                conn_prop_data["min"][sidx, tidx, pidx] = np.min(prop_values).astype(np.float64)
                conn_prop_data["max"][sidx, tidx, pidx] = np.max(prop_values).astype(np.float64)
                hist_counts, bin_edges = np.histogram(prop_values, bins=hist_bins)
                conn_prop_data["hist"][sidx, tidx, pidx] = (hist_counts, bin_edges)
                bin_centers = [np.mean(bin_edges[i : i + 2]) for i in range(len(hist_counts))]
                loc, scale, _ = _norm_fitting(
                    bin_centers,
                    hist_counts,
                    def_mn=conn_prop_data["mean"][sidx, tidx, pidx],
                    def_sd=conn_prop_data["std"][sidx, tidx, pidx],
                )
                conn_prop_data["norm_loc"][sidx, tidx, pidx] = loc
                conn_prop_data["norm_scale"][sidx, tidx, pidx] = scale
                v, c = np.unique(prop_values, return_counts=True)
                if len(v) == 1 or (
                    len(v) < len(prop_values) and len(v) <= MAX_UNIQUE_COUNT
                ):  # Store discrete values only if not too many [otherwise, it is probably not a discrete distribution]
                    conn_prop_data["val"][sidx, tidx, pidx] = v
                    conn_prop_data["cnt"][sidx, tidx, pidx] = c
                    conn_prop_data["p"][sidx, tidx, pidx] = c / np.sum(c)
                else:
                    conn_prop_data["val"][sidx, tidx, pidx] = []
                    conn_prop_data["cnt"][sidx, tidx, pidx] = []
                    conn_prop_data["p"][sidx, tidx, pidx] = []
                    # log.warning(f'Discrete {p} values not stored for {m_types[0][sidx]}-{m_types[1][tidx]} ({len(v)} of {len(prop_values)} unique values)!')

    log.debug(
        f'Between {conn_counts["min"]} and {conn_counts["max"]} connections per pathway found. {conn_counts["sel"]} of {len(m_types[0])}x{len(m_types[1])} pathways selected.'
    )

    return {
        "syns_per_conn_data": syns_per_conn_data,
        "conn_prop_data": conn_prop_data,
        "m_types": m_types,
        "m_type_class": m_type_class,
        "m_type_layer": m_type_layer,
        "syn_props": syn_props,
        "hist_bins": hist_bins,
    }


def build(
    syns_per_conn_data,
    conn_prop_data,
    m_types,
    m_type_class,
    m_type_layer,
    syn_props,
    distr_types={},
    data_types={},
    data_bounds={},
    shared_within={},
    **_,
):
    """Fit model distribution to data, incl. missing values interpolated at different levels of granularity.

    Args:
        syns_per_conn_data (dict): Dictionary with entries for all statistics (see Notes under :func:`extract`) estimated for #synapses/connection, each of which contains a numpy.ndarray of size <#source-mtypes x #target-mtypes> for all pairs of m-types, as returned by :func:`extract`
        conn_prop_data (dict): Dictionary with entries for all statistics (see Notes under :func:`extract`) estimated for all synaptic properties, each of which contains a numpy.ndarray of size <#source-mtypes x #target-mtypes x #properties> for all pairs of m-types and synapse properties, as returned by :func:`extract`
        m_types (list): Two-element list of lists of source (pre-synaptic) and target (post-synaptic) m-types, as returned by :func:`extract`
        m_type_class (list): Two-element list of lists of synapse classes (i.e., EXC, INH) belonging to each source and target m-type (assuming that each m-type corresponds to exactly one synapse class), as returned by :func:`extract`
        m_type_layer (list): Two-element list of lists of layers belonging to each source and target m-type (assuming that each m-type corresponds to exactly one cortical layer), as returned by :func:`extract`
        syn_props (list-like): List of synaptic property names stored in ``conn_prop_data``, as returned by :func:`extract`
        distr_types (dict): Optional dictionary specifying the distribution type (dict value) for each property (dict key); if omitted, a "normal" distribution is assumed (and will raise a warning); see Notes for available distribution types
        data_types (dict): Optional dictionary specifying the output data type (dict value; e.g., "int", "float", ...) for each property (dict key) when drawing values from the fitted model
        data_bounds (dict): Optional dictionary specifying the output data bounds (dict value; list-like with two elements for lower/upper bounds) for each property (dict key) when drawing values from the fitted model
        shared_within (dict): Optional dictionary specifying if the same values are shared among synapses belonging to the same connections (boolean dict value) for each property (dict key) when drawing values from the fitted model; can be used to manually overwrite the data-derived value

    Returns:
        connectome_manipulator.model_building.model_types.ConnPropsModel: Fitted connection/synapse properties model

    Note:
        The property name "n_syn_per_conn" (defined by ``model_types.N_SYN_PER_CONN_NAME``) can be used in ``distr_types``, ``data_types``, and ``data_bounds`` dicts to specify distribution types, data types, and bounds for the #synapses/connection property.

    Note:
        The following distribution types are supported:

        * "constant": Constant value (define by "mean")
        * "normal": Gaussian normal distribution (define by "mean", "std")
        * "truncnorm": Truncated normal distribution (define by "norm_loc", "norm_scale", "min", "max")
        * "gamma": Gamma distribution (define by "mean", "std")
        * "poisson": Poisson distribution (define by "mean")
        * "ztpoisson": Zero-truncated poisson distribution (define by "mean")
        * "discrete": Discrete distribution (define by "val", "p")
        * "zero": Empty distribution always returning zero; can be used to model unused parameters
    """
    # Interpolate missing values in lookup tables
    syns_per_conn_model = {k: v.copy() for (k, v) in syns_per_conn_data.items()}
    conn_prop_model = {k: v.copy() for (k, v) in conn_prop_data.items()}
    missing_list = np.array(np.where(np.logical_not(np.isfinite(syns_per_conn_model["mean"])))).T
    level_counts = {}  # Count interpolation levels for reporting
    for sidx, tidx in missing_list:
        # Select level of granularity
        for level in range(5):
            if level == 0:  # Use source m-type/target layer/synapse class value, if existent
                src_sel = np.array([sidx])
                tgt_sel = np.where(
                    np.logical_and(
                        np.array(m_type_layer[1]) == m_type_layer[1][tidx],
                        np.array(m_type_class[1]) == m_type_class[1][tidx],
                    )
                )[0]
            elif level == 1:  # Use source m-type/target synapse class value, if existent
                src_sel = np.array([sidx])
                tgt_sel = np.where(np.array(m_type_class[1]) == m_type_class[1][tidx])[0]
            elif level == 2:  # Use per layer/synapse class value, if existent
                src_sel = np.where(
                    np.logical_and(
                        np.array(m_type_layer[0]) == m_type_layer[0][sidx],
                        np.array(m_type_class[0]) == m_type_class[0][sidx],
                    )
                )[0]
                tgt_sel = np.where(
                    np.logical_and(
                        np.array(m_type_layer[1]) == m_type_layer[1][tidx],
                        np.array(m_type_class[1]) == m_type_class[1][tidx],
                    )
                )[0]
            elif level == 3:  # Use per synapse class value, if existent
                src_sel = np.where(np.array(m_type_class[0]) == m_type_class[0][sidx])[0]
                tgt_sel = np.where(np.array(m_type_class[1]) == m_type_class[1][tidx])[0]
            else:  # Otherwise: Use overall value
                src_sel = np.array(list(range(len(m_types[0]))))
                tgt_sel = np.array(list(range(len(m_types[1]))))
            if np.any(np.isfinite(syns_per_conn_data["mean"][src_sel, :][:, tgt_sel])):
                level_counts[f"Level{level}"] = level_counts.get(f"Level{level}", 0) + 1
                break

        def merge_uniq(vals, cnts):
            """Helper function to merge unique values/counts"""
            vals = list(
                filter(lambda x: len(x) > 0 if hasattr(x, "__iter__") else np.isfinite(x), vals)
            )  # [Remove NaNs/empty lists, so not to mess up data type]
            cnts = list(
                filter(lambda x: len(x) > 0 if hasattr(x, "__iter__") else np.isfinite(x), cnts)
            )  # [Remove NaNs/empty lists, so not to mess up data type]
            if len(vals) > 0:
                vc_dict = {
                    v: 0 for v in np.unique(np.hstack(vals))
                }  # Init value/count dict [increasing order!]
                for v, c in zip(np.hstack(vals), np.hstack(cnts)):
                    vc_dict[v] += c  # Add to existing count
                vals = np.array(list(vc_dict.keys()))
                cnts = np.array([vc_dict[v] for v in vals])
            return vals, cnts

        # Interpolate missing values
        syns_per_conn_model["mean"][sidx, tidx] = np.nanmean(
            syns_per_conn_data["mean"][src_sel, :][:, tgt_sel]
        )
        syns_per_conn_model["std"][sidx, tidx] = np.nanmean(
            syns_per_conn_data["std"][src_sel, :][:, tgt_sel]
        )
        syns_per_conn_model["min"][sidx, tidx] = np.nanmin(
            syns_per_conn_data["min"][src_sel, :][:, tgt_sel]
        )
        syns_per_conn_model["max"][sidx, tidx] = np.nanmax(
            syns_per_conn_data["max"][src_sel, :][:, tgt_sel]
        )
        syns_per_conn_model["norm_loc"][sidx, tidx] = np.nanmean(
            syns_per_conn_data["norm_loc"][src_sel, :][:, tgt_sel]
        )
        syns_per_conn_model["norm_scale"][sidx, tidx] = np.nanmean(
            syns_per_conn_data["norm_scale"][src_sel, :][:, tgt_sel]
        )

        uvals = syns_per_conn_data["val"][src_sel, :][:, tgt_sel].flatten()
        ucnts = syns_per_conn_data["cnt"][src_sel, :][:, tgt_sel].flatten()
        v, c = merge_uniq(uvals, ucnts)
        syns_per_conn_model["val"][sidx, tidx] = v
        syns_per_conn_model["cnt"][sidx, tidx] = c
        syns_per_conn_model["p"][sidx, tidx] = c / np.sum(c)

        conn_prop_model["mean"][sidx, tidx, :] = [
            np.nanmean(conn_prop_data["mean"][src_sel, :, p][:, tgt_sel])
            for p in range(len(syn_props))
        ]
        conn_prop_model["std"][sidx, tidx, :] = [
            np.nanmean(conn_prop_data["std"][src_sel, :, p][:, tgt_sel])
            for p in range(len(syn_props))
        ]
        conn_prop_model["min"][sidx, tidx, :] = [
            np.nanmin(conn_prop_data["min"][src_sel, :, p][:, tgt_sel])
            for p in range(len(syn_props))
        ]
        conn_prop_model["max"][sidx, tidx, :] = [
            np.nanmax(conn_prop_data["max"][src_sel, :, p][:, tgt_sel])
            for p in range(len(syn_props))
        ]
        conn_prop_model["norm_loc"][sidx, tidx, :] = [
            np.nanmean(conn_prop_data["norm_loc"][src_sel, :, p][:, tgt_sel])
            for p in range(len(syn_props))
        ]
        conn_prop_model["norm_scale"][sidx, tidx, :] = [
            np.nanmean(conn_prop_data["norm_scale"][src_sel, :, p][:, tgt_sel])
            for p in range(len(syn_props))
        ]
        conn_prop_model["shared_within"][sidx, tidx, :] = [
            np.round(np.nanmean(conn_prop_data["shared_within"][src_sel, :, p][:, tgt_sel])).astype(
                bool
            )
            for p in range(len(syn_props))
        ]  # Majority vote in case of inconsistent sharing behavior

        for pidx in range(len(syn_props)):
            uvals = [conn_prop_model["val"][s][t][pidx] for s in src_sel for t in tgt_sel]
            ucnts = [conn_prop_model["cnt"][s][t][pidx] for s in src_sel for t in tgt_sel]
            v, c = merge_uniq(uvals, ucnts)
            conn_prop_model["val"][sidx, tidx, pidx] = v
            conn_prop_model["cnt"][sidx, tidx, pidx] = c
            conn_prop_model["p"][sidx, tidx, pidx] = c / np.sum(c)

    log.info(
        f"Interpolated {missing_list.shape[0]} missing values. Interpolation level counts: {{k: level_counts[k] for k in sorted(level_counts.keys())}}"
    )

    # Create model properties dictionary
    prop_model_dict = {}
    for pidx, prop in enumerate(syn_props + [model_types.N_SYN_PER_CONN_NAME]):
        prop_model_dict[prop] = {}
        if prop not in distr_types:
            log.warning(f'No distribution type for "{prop}" specified - Using "normal"!')
        distr_type = distr_types.get(prop, "normal")
        log.log_assert(
            distr_type in model_types.ConnPropsModel.distribution_attributes,
            f'ERROR: Distribution type "{distr_type}" not supported!',
        )
        dtype = data_types.get(prop)
        bounds = data_bounds.get(prop)
        shared = shared_within.get(prop)
        for sidx, src in enumerate(m_types[0]):
            prop_model_dict[prop][src] = {}
            for tidx, tgt in enumerate(m_types[1]):
                attr_dict = {"type": distr_type}
                distr_attr = model_types.ConnPropsModel.distribution_attributes[distr_type]
                if prop == model_types.N_SYN_PER_CONN_NAME:
                    log.log_assert(
                        np.all([attr in syns_per_conn_model for attr in distr_attr]),
                        f'ERROR: Not all required attribute(s) {distr_attr} for distribution "{distr_type}" found!',
                    )
                    attr_dict.update(
                        {attr: syns_per_conn_model[attr][sidx, tidx] for attr in distr_attr}
                    )
                else:
                    distr_attr = distr_attr + ["shared_within"]
                    log.log_assert(
                        np.all([attr in conn_prop_model for attr in distr_attr]),
                        f'ERROR: Not all required attribute(s) {distr_attr} for distribution "{distr_type}" found!',
                    )
                    attr_dict.update(
                        {attr: conn_prop_model[attr][sidx, tidx, pidx] for attr in distr_attr}
                    )
                if dtype is not None:
                    attr_dict.update({"dtype": dtype})
                if bounds is not None and hasattr(bounds, "__iter__") and len(bounds) == 2:
                    if bounds[0] is not None:
                        attr_dict.update({"lower_bound": bounds[0]})
                    if bounds[1] is not None:
                        attr_dict.update({"upper_bound": bounds[1]})
                if shared is not None and isinstance(shared, bool):
                    # Overwrite data-derived value
                    attr_dict.update({"shared_within": shared})
                elif distr_type == "zero":
                    # Overwrite data-derived value for unused properties
                    attr_dict.update({"shared_within": False})
                prop_model_dict[prop][src][tgt] = attr_dict

    # Create model
    model = model_types.ConnPropsModel(
        src_types=m_types[0], tgt_types=m_types[1], prop_stats=prop_model_dict
    )
    log.debug("Model description:\n%s", model)

    return model


def plot(
    out_dir,
    syns_per_conn_data,
    conn_prop_data,
    m_types,
    syn_props,
    model,
    plot_sample_size=1000,
    **_,
):  # pragma: no cover
    """Visualizes extracted data vs. actual model output.

    Args:
        out_dir (str): Path to output directory where the results figures will be stored
        syns_per_conn_data (dict): Dictionary with entries for all statistics (see Notes under :func:`extract`) estimated for #synapses/connection, each of which contains a numpy.ndarray of size <#source-mtypes x #target-mtypes> for all pairs of m-types, as returned by :func:`extract`
        conn_prop_data (dict): Dictionary with entries for all statistics (see Notes under :func:`extract`) estimated for all synaptic properties, each of which contains a numpy.ndarray of size <#source-mtypes x #target-mtypes x #properties> for all pairs of m-types and synapse properties, as returned by :func:`extract`
        m_types (list): Two-element list of lists of source (pre-synaptic) and target (post-synaptic) m-types, as returned by :func:`extract`
        syn_props (list-like): List of synaptic property names stored in ``conn_prop_data``, as returned by :func:`extract`
        model (connectome_manipulator.model_building.model_types.ConnPropsModel): Fitted connection/synapse properties model, as returned by :func:`build`
        plot_sample_size (int): Number of samples to draw when plotting model distributions
    """
    model_params = model.get_param_dict()
    prop_names = model.get_prop_names()

    # Plot data vs. model: property maps
    title_str = ["Data", "Model"]
    for stat_sel in ["mean", "std", "shared_within"]:
        for pidx, p in enumerate(prop_names):
            if pidx < len(prop_names) - 1:
                data_stat_sel = conn_prop_data[stat_sel][:, :, pidx]
            else:  # Last element is n_syn_per_conn
                if stat_sel not in syns_per_conn_data:
                    continue  # Skip if statistics does not exist here
                data_stat_sel = syns_per_conn_data[stat_sel]
            plt.figure(figsize=(8, 3), dpi=300)
            model_stat_sel = np.full((len(m_types[0]), len(m_types[1])), np.nan)
            for sidx, s in enumerate(m_types[0]):
                for tidx, t in enumerate(m_types[1]):
                    model_stat_sel[sidx, tidx] = _get_model_stat(
                        stat_sel, model_params["prop_stats"][p][s][t]
                    )
            for didx, data in enumerate([data_stat_sel, model_stat_sel]):
                plt.subplot(1, 2, didx + 1)
                plt.imshow(data, interpolation="nearest", cmap="jet")
                plt.xticks(range(len(m_types[1])), m_types[1], rotation=90, fontsize=3)
                plt.yticks(range(len(m_types[0])), m_types[0], rotation=0, fontsize=3)
                plt.colorbar()
                plt.title(title_str[didx])
            plt.suptitle(f"{p} ({stat_sel})", fontweight="bold")
            plt.tight_layout()

            out_fn = os.path.abspath(
                os.path.join(out_dir, f"data_vs_model_map_{stat_sel}__{p}.png")
            )
            log.info(f"Saving {out_fn}...")
            plt.savefig(out_fn)

    # Plot data vs. model: Distribution histogram examples (generative model) + AUC
    conn_counts = [
        [
            (
                np.sum(syns_per_conn_data["hist"][sidx, tidx][0])
                if (
                    hasattr(syns_per_conn_data["hist"][sidx, tidx], "__iter__")
                    and len(syns_per_conn_data["hist"][sidx, tidx]) > 0
                )
                else 0
            )
            for tidx in range(len(m_types[1]))
        ]
        for sidx in range(len(m_types[0]))
    ]
    max_pathways = np.where(
        np.array(conn_counts) == np.max(conn_counts)
    )  # Select pathway(s) with maximum number of connections (i.e., most robust statistics)
    sidx, tidx = [
        max_pathways[0][0],
        max_pathways[1][0],
    ]  # Select first of these pathways for plotting
    src, tgt = [m_types[0][sidx], m_types[1][tidx]]
    for pidx, p in enumerate(prop_names):
        plt.figure(figsize=(5, 3), dpi=300)
        if pidx < len(syn_props):
            data_hist = conn_prop_data["hist"][sidx, tidx, pidx]
        else:
            data_hist = syns_per_conn_data["hist"][sidx, tidx]
        plt.bar(
            data_hist[1][:-1],
            data_hist[0] / np.sum(data_hist[0]),
            align="edge",
            width=np.min(np.diff(data_hist[1])),
            label=f"Data (N={np.max(conn_counts)})",
        )
        model_data = np.hstack(
            [
                model.draw(prop_name=p, src_type=src, tgt_type=tgt, size=1)
                for n in range(plot_sample_size)
            ]
        )  # Draw <plot_sample_size> single values from property distribution
        hist_bins = data_hist[1]  # Use same model distribution binning as for data
        bin_size = np.min(np.diff(hist_bins))
        if min(model_data) < hist_bins[0]:  # Extend binning to lower values to cover whole range
            hist_bins = np.hstack(
                [
                    np.flip(np.arange(hist_bins[0], min(model_data) - bin_size, -bin_size)),
                    hist_bins[1:],
                ]
            )
        if max(model_data) > hist_bins[-1]:  # Extend binning to higher values to cover whole range
            hist_bins = np.hstack(
                [hist_bins[:-1], np.arange(hist_bins[-1], max(model_data) + bin_size, bin_size)]
            )
        model_hist = np.histogram(model_data, bins=hist_bins)
        plt.step(
            model_hist[1],
            np.hstack([model_hist[0][0], model_hist[0]]) / np.sum(model_hist[0]),
            where="pre",
            color="tab:orange",
            label=f"Model (N={plot_sample_size})",
        )
        plt.grid()
        plt.gca().set_axisbelow(True)
        plt.title(f"{src} to {tgt}", fontweight="bold")
        plt.xlabel(p)
        plt.ylabel("Density")
        plt.legend()
        plt.tight_layout()

        out_fn = os.path.abspath(os.path.join(out_dir, f"data_vs_model_hist__{p}.png"))
        log.info(f"Saving {out_fn}...")
        plt.savefig(out_fn)

        # AUC/ROC plot
        (
            AUC,
            ERR,
            _FPR,
            _TPR,
            _distr_norm_A,
            _distr_norm_B,
            _pos_norm_A,
            _pos_norm_B,
            xp,
            yp,
            x,
            y,
        ) = _compute_AUC(data_hist[0], model_hist[0], data_hist[1], model_hist[1])

        plt.figure()
        plt.plot([0.0, 1.0], [0.0, 1.0], "--k")
        plt.plot(xp, yp, ".-")
        plt.plot(x, y, ".--")
        plt.grid()
        plt.gca().set_axisbelow(True)
        plt.title(f"{src} to {tgt}: {p} (AUC={AUC:.2f}, ERR={ERR:.2f})", fontweight="bold")
        plt.xlabel("FPR")
        plt.ylabel("TPR")
        plt.tight_layout()

        out_fn = os.path.abspath(os.path.join(out_dir, f"data_vs_model_AUC__{p}.png"))
        log.info(f"Saving {out_fn}...")
        plt.savefig(out_fn)


# Helper functions
def _norm_fitting(
    hist_values,
    hist_counts,
    max_nfev=30,
    rel_fit_err_th=0.5,
    def_mn=np.nan,
    def_sd=np.nan,
    def_sc=np.nan,
):
    """Helper function to extract fitted norm mean/SD/scaling attributes from histogram (using a large error threshold by default in order to get at least a rough estimate)."""

    def norm_fct(x, mn, sd, sc):
        return sc * norm(loc=mn, scale=sd).pdf(x)

    p0 = [np.mean(hist_values), 1.0, 1.0]
    bounds = [[min(hist_values), 0, 0], [max(hist_values), np.inf, np.inf]]
    # Note: "mean" assumed to be within hist value bounds!!

    invalid_fit = False
    try:
        (mn_opt, sd_opt, sc_opt), pcov, *_ = curve_fit(
            norm_fct, hist_values, hist_counts, p0=p0, bounds=bounds, max_nfev=max_nfev
        )
    except (
        ValueError,
        RuntimeError,
    ):  # Raised if input data invalid or optimization fails
        invalid_fit = True

    if not invalid_fit:
        rel_err = np.sqrt(np.diag(pcov)) / np.array(
            [mn_opt, sd_opt, sc_opt]
        )  # Rel. standard error of the coefficients
        # log.debug(f"Rel. error of norm model fit: {rel_err}")
        if not all(np.isfinite(rel_err)) or max(rel_err) > rel_fit_err_th:
            # log.error(
            #     f"Rel. error of norm model fit exceeds error threshold of {rel_fit_err_th} (or could not be determined)!"
            # )
            invalid_fit = True

    if invalid_fit:  # Set default values
        mn_opt = def_mn
        sd_opt = def_sd
        sc_opt = def_sc

    return mn_opt, sd_opt, sc_opt


def _check_shared_property(edges_sel, prop_name, conn_sel, syn_conn_idx):
    """Check if shared property values within connections."""
    if len(conn_sel) > 1 and len(np.unique(edges_sel[prop_name])) == 1:
        # In case of a constant overall distribution, assume no sharing
        is_shared = False
    else:
        is_shared = True
        for c in conn_sel:
            if len(np.unique(edges_sel.loc[syn_conn_idx == c, prop_name])) > 1:
                # Found different property values within same connection
                is_shared = False
                break
    return is_shared


def _get_property_values(edges_sel, prop_name, conn_sel, syn_conn_idx, is_shared):
    """Collect property values over connections (if shared values) or synapses (if non-shared values)."""
    prop_values = []
    for c in conn_sel:
        if is_shared:
            # Shared within connection, so take only first value
            prop_values.append(edges_sel.loc[syn_conn_idx == c, prop_name].iloc[0])
        else:
            # Different values within connection, so take all values
            prop_values.append(edges_sel.loc[syn_conn_idx == c, prop_name].to_numpy())
    prop_values = np.hstack(prop_values)
    return prop_values


def _compute_AUC(distr_A, distr_B, bins_A, bins_B, dth=0.05, dx=0.01):
    """Computes area under the ROC curve for comparing two distributions."""
    pos_A = np.array(
        [np.mean(bins_A[i : i + 2]) for i in range(len(bins_A) - 1)]
    )  # Bin center positions
    pos_B = np.array(
        [np.mean(bins_B[i : i + 2]) for i in range(len(bins_B) - 1)]
    )  # Bin center positions

    min_range = np.minimum(pos_A[0], pos_B[0])
    max_range = np.maximum(pos_A[-1], pos_B[-1])
    pos_norm_A = (pos_A - min_range) / (max_range - min_range)  # Normalized positions
    pos_norm_B = (pos_B - min_range) / (max_range - min_range)  # Normalized positions

    bin_size_A = np.mean(np.diff(bins_A))
    bin_size_B = np.mean(np.diff(bins_B))
    distr_norm_A = distr_A / (np.sum(distr_A) * bin_size_A)
    distr_norm_B = distr_B / (np.sum(distr_B) * bin_size_B)

    ths = np.arange(0, 1.0 + dth, dth)  # Thresholds
    TPR = np.array([np.sum(distr_norm_A[pos_norm_A >= th] * bin_size_A) for th in ths])
    FPR = np.array([np.sum(distr_norm_B[pos_norm_B >= th] * bin_size_B) for th in ths])

    sort_idx = np.argsort(FPR)
    xp = FPR[sort_idx]
    yp = TPR[sort_idx]

    x = np.arange(0.0, 1.0 + dx, dx)
    y = np.interp(x, xp, yp)
    AUC = np.trapz(y, x, dx) - 0.5
    ERR = np.trapz(np.abs(y - x), x, dx)  # Error: Area of abs. differences

    return AUC, ERR, FPR, TPR, distr_norm_A, distr_norm_B, pos_norm_A, pos_norm_B, xp, yp, x, y


def _get_model_stat(stat, m_params):
    """Get distribution statistic (if existing) or derive from other model paramters, if possible."""
    val = np.nan
    if stat in m_params:  # Return existing stat. parameter
        val = m_params[stat]
    else:
        if m_params["type"] == "constant":
            if stat == "std":
                val = 0.0
            elif stat in {"min", "max"}:
                val = m_params["mean"]
        elif m_params["type"] == "discrete":
            # Derive missing statistics from discrete values/probabilities
            if stat == "mean":
                val = np.sum(np.array(m_params["p"] * np.array(m_params["val"])))
            elif stat == "std":
                m = np.sum(np.array(m_params["p"] * np.array(m_params["val"])))
                val = np.sqrt(
                    np.sum(np.array(m_params["p"]) * (np.array(m_params["val"]) - m) ** 2)
                )
            elif stat == "min":
                val = np.min(np.array(m_params["val"])[np.array(m_params["p"]) > 0.0])
            elif stat == "max":
                val = np.max(np.array(m_params["val"])[np.array(m_params["p"]) > 0.0])
        elif m_params["type"] == "truncnorm":
            # Estimate missing statistics from generated truncnorm distribution
            distr = model_types.ConnPropsModel.draw_from_distribution(m_params, size=200)
            if stat == "mean":
                val = np.mean(distr)
            elif stat == "std":
                val = np.std(distr)
        elif m_params["type"] == "poisson":
            if stat == "std":
                # Derive std from poisson mean
                val = np.sqrt(m_params["mean"])
            elif stat == "min":
                val = 0
        elif m_params["type"] == "ztpoisson":
            if stat == "std":
                # Derive std from zero-truncated poisson mean
                mn = m_params["mean"]
                lam = model_types.ConnPropsModel.compute_ztpoisson_lambda(mn)
                val = np.sqrt(mn * (1 + lam - mn))
            elif stat == "min":
                val = 1
        elif m_params["type"] == "zero":
            # Set missing statistics to zero
            val = 0.0
    return val
